# For legacy code, include these modules in the top-level analysis module
# since they existed in base before becoming sub-modules
from .base import *
from .neuropixel import *
from .quality import *
from .wrappers import *

# For new submodules, import just the namespace
from . import optitrack
from . import bmi3d


from matplotlib.animation import FuncAnimation
import os
import matplotlib.pyplot as plt
import numpy as np

from .base import plot_spatial_map, plot_targets, set_bounds
from .. import postproc

def saveanim(animation, base_dir, filename, dpi=100, **savefig_kwargs):
    '''
    Save an animation using ffmpeg

    Args:
        animation (pyplot.Animation): animation to save
        base_dir (str): directory to write
        filename (str): should end in '.mp4'
        dpi (float): resolution of the video file
        savefig_kwargs (kwargs, optional): arguments to pass to savefig
    '''
    filepath = os.path.join(base_dir, filename)
    animation.save(filepath, dpi=dpi, savefig_kwargs=savefig_kwargs)


def showanim(animation, closeanim=True):
    '''
    Display an animation in a python notebook

    Args:
        animation (pyplot.Animation): animation to display
        closeanim (bool, optional): also close the animation figure to avoid showing a static plot
    '''
    from IPython import display # not a required package
    html = display.HTML(animation.to_html5_video())
    display.display(html)
    if closeanim:
        plt.close()

def animate_events(events, times, fps, xy=(0.3, 0.3), fontsize=30, color='g'):
    '''
    Silly function to plot events as text, frame by frame in an animation

    Args:
        events (list): list of event names or numbers
        times (list): timestamps of each event
        fps (float): sampling rate to animate
        xy (tuple, optional): (x, y) coorindates of the left bottom corner of each event label, from 0 to 1.
        fontsize (float, optional): size to draw the event labels

    Returns:
        matplotlib.animation.FuncAnimation: animation object

    Example:
        .. raw:: html

            <video controls src="_static/test_anim_events.mp4"></video>
    '''
    frame_events, event_names = postproc.sample_events(events, times, fps)

    def display_text(num, events, names, note):
        display = names[events[num, :] == 1]
        if len(display) > 0:
            note.set_text(display[0])  # note if simultaneous events occur, we just print the first

    fig, ax = plt.subplots(1, 1)
    note = ax.annotate("", xy, fontsize=fontsize, color=color)
    plt.axis('off')
    return FuncAnimation(fig, display_text, frames=frame_events.shape[0],
                         interval=round(1000 / fps),
                         fargs=(frame_events, event_names, note))


def animate_trajectory_3d(trajectory, samplerate, history=1000, color='b',
                          axis_labels=['x', 'y', 'z']):
    '''
    Draws a trajectory moving through 3D space at the given sampling rate and with a
    fixed maximum number of points visible at a time.

    Args:
        trajectory (n, 3): matrix of n points
        samplerate (float): sampling rate of the trajectory data
        history (int, optional): maximum number of points visible at once
    
    Returns:
        matplotlib.animation.FuncAnimation: animation object

    Example:
        .. raw:: html

            <video controls src="_static/test_anim_trajectory.mp4"></video>    
    '''

    fig = plt.figure()
    ax = plt.subplot(111, projection='3d')

    line, = ax.plot(trajectory[0, 0], trajectory[0, 1], trajectory[0, 2], color=color)

    ax.set_xlim((np.nanmin(trajectory[:, 0]), np.nanmax(trajectory[:, 0])))
    ax.set_xlabel(axis_labels[0])

    ax.set_ylim((np.nanmin(trajectory[:, 1]), np.nanmax(trajectory[:, 1])))
    ax.set_ylabel(axis_labels[1])

    ax.set_zlim((np.nanmin(trajectory[:, 2]), np.nanmax(trajectory[:, 2])))
    ax.set_zlabel(axis_labels[2])

    def draw(num):
        length = min(num, history)
        start = num - length
        line.set_data(trajectory[start:num, 0], trajectory[start:num, 1])
        line.set_3d_properties(trajectory[start:num, 2])
        return line,

    return FuncAnimation(fig, draw, frames=trajectory.shape[0],
                         init_func=lambda: None, interval=1000. / samplerate)

def animate_spatial_map(data_map, x, y, samplerate, cmap='bwr'):
    '''
    Animates a 2d heatmap. Use :func:`aopy.visualization.get_data_map` to get a 2d array
    for each timepoint you want to animate, then put them into a list and feed them to this
    function. See also :func:`aopy.visualization.show_anim` and :func:`aopy.visualization.save_anim`

    Example:
        ::
        
            samplerate = 20
            duration = 5
            x_pos, y_pos = np.meshgrid(np.arange(0.5,10.5),np.arange(0.5, 10.5))
            data_map = []
            for frame in range(duration*samplerate):
                t = np.linspace(-1, 1, 100) + float(frame)/samplerate
                c = np.sin(t)
                data_map.append(get_data_map(c, x_pos.reshape(-1), y_pos.reshape(-1)))

            filename = 'spatial_map_animation.mp4'
            ani = animate_spatial_map(data_map, x_pos, y_pos, samplerate, cmap='bwr')
            saveanim(ani, write_dir, filename)

        .. raw:: html

            <video controls src="_static/spatial_map_animation.mp4"></video>

    Args:
        data_map (nt): array of 2d maps
        x (list): list of x positions
        y (list): list of y positions
        samplerate (float): rate of the data_map samples
        cmap (str, optional): name of the colormap to use. Defaults to 'bwr'.
    
    Returns:
        matplotlib.animation.FuncAnimation: animation object
    '''

    # Plotting subroutine
    def plotdata(i):
        im.set_data(data_map[i])
        return im

    # Initial plot
    fig, ax = plt.subplots()
    im = plot_spatial_map(data_map[0], x, y, ax=ax, cmap=cmap)

    # Change the color limits
    min_c = np.min(np.array(data_map))
    max_c = np.max(np.array(data_map))
    im.set_clim(min_c, max_c)
        
    # Create animation
    ani = FuncAnimation(fig, plotdata, frames=len(data_map),
                            interval=1000./samplerate)

    return ani

def animate_cursor_eye(cursor_trajectory, eye_trajectory, samplerate, target_positions, target_radius, 
                       bounds, cursor_radius=0.5, eye_radius=0.25,
                       cursor_color='blue', eye_color='purple'):
    '''
    Draws an animation of two trajectories with static targets. The colors and endpoint radii of the
    two trajectories can be specified along with the position and radius of the targets. Targets are
    colored automatically according to :func:`~aopy.visualization.plot_targets`.

    Example:

        .. raw:: html

            <video controls src="_static/test_anim_cursor_eye.mp4"></video>

    Args:
        cursor_trajectory ((nt, ndim) array): Cursor positions over time for 2D or 3D trajectories.
        eye_trajectory ((nt, ndim) array): Eye positions over time for 2D or 3D trajectories.
        samplerate (float): The sampling rate of the trajectories in Hz.
        target_positions ((ntargets, ndim) array): Array of target positions for 2D or 3D targets.
        target_radius (float): Radius of the targets.
        bounds (tuple): Boundaries of the plot area. See :func:`~aopy.visualization.plot_targets`.
        cursor_radius (float, optional): Radius of the cursor endpoint. Default is 0.5.
        eye_radius (float, optional): Radius of the eye endpoint. Default is 0.25.
        cursor_color (plt.color, optional): Color of the cursor trajectory. Default is 'blue'.
        eye_color (plt.color, optional): Color of the eye trajectory. Default is 'purple'.

    Returns:
        None
    
    Returns:
        matplotlib.animation.FuncAnimation: animation object
    '''
    assert len(cursor_trajectory) == len(eye_trajectory), "Cursor and Eye trajectories must have the same length"

    def plotdata(i):
        cur.center = cursor_trajectory[i]
        eye.center = eye_trajectory[i]
        cur_line.set_data(*cursor_trajectory[:i+1].T)
        eye_line.set_data(*eye_trajectory[:i+1].T)
        return ax
            
    # # Initial plot
    fig, ax = plt.subplots(1, 1)
    plot_targets(target_positions, target_radius, bounds=bounds, ax=ax)
    cur = plt.Circle(cursor_trajectory[0], radius=cursor_radius, alpha=0.5, color=cursor_color)
    eye = plt.Circle(eye_trajectory[0], radius=eye_radius, alpha=0.5, color=eye_color)
    ax.add_artist(cur)
    ax.add_artist(eye)
    cur_line, = plt.plot(*cursor_trajectory[:1].T, color=cursor_color)
    eye_line, = plt.plot(*eye_trajectory[:1].T, color=eye_color)

    # Create animation
    ani = FuncAnimation(fig, plotdata, 
                        frames=len(cursor_trajectory), interval=1000./samplerate)  
    return ani

def get_animate_circles_func(samplerate, bounds, circle_radii, circle_colors, *circle_ts, history=1., ax=None):
    '''
    Draws an animation of an arbitrary number of circles. Used in :func:`~aopy.visualization.animation.animate_behavior`.

    Args:
        samplerate (float): The sampling rate of the trajectories in Hz.
        bounds (tuple): Boundaries of the plot area. See :func:`~aopy.visualization.plot_targets`.
        circle_radii (list of float): Radius of each circle.
        circle_colors (list of plt.color): Color of each circle.
        circle_ts (list of (nt, 2) arrays): Circle positions over time for 2D trajectories.
        history (float, optional): how long (in seconds) to animate lines trailing the circles. Default 1.
        ax (pyplot.Axes, optional): axis on which to plot the animation

    Returns:
        function: plotting function for FuncAnimation
    '''
    ncircles = len(circle_ts)
    nhist = int(history*samplerate)
            
    # Initial plot
    if ax is None:
        ax = plt.gca()
    set_bounds(bounds, ax=ax)
    ax.set_aspect('equal', adjustable='box')
    circles = []
    lines = []
    for j in range(ncircles):
        circles.append(plt.Circle(circle_ts[j][0], radius=circle_radii[j], alpha=0.5, color=circle_colors[j]))
        ax.add_artist(circles[-1])
        lines.append(plt.plot(*(circle_ts[j][:1,:2].T), color=circle_colors[j])[0])
    
    # Plotting function
    def plotdata(i):
        for j in range(ncircles):
            circles[j].center = circle_ts[j][i]
            lines[j].set_data(*circle_ts[j][max(0,i-nhist):i+1,:2].T)
        return ax

    return plotdata
                       
                       
def animate_behavior(targets, cursor, eye, samplerate, bounds, 
                     target_radius, target_colors, cursor_radius, cursor_color='blue', 
                     eye_radius=0.25, eye_color='purple', history=0.):
    '''
    Animate target, cursor, and eye data together. 

    Args:
        targets (list of (nt,) arrays): Target position timeseires for each target.
        cursor ((nt, 2) array): Cursor position timeseires.
        eye ((nt, 2) array): Eye position timeseires.
        samplerate (float): The sampling rate of all the trajectories in Hz.
        bounds (tuple): Boundaries of the plot area. See :func:`~aopy.visualization.plot_targets`.
        target_radius (float): Radius of the targets.
        target_colors (list of plt.color): Color of each target.
        cursor_radius (float): Radius of the cursor.
        cursor_color (plt.color, optional): Color of the cursor. Default is 'blue'.
        eye_radius (float): Radius of the eye circle.
        eye_color (plt.color, optional): Color of the eye trajectory. Default is 'purple'.
        history (float, optional): how long (in seconds) to animate lines trailing the circles. Default 0.

    Returns:
        matplotlib.animation.FuncAnimation: animation object

    Example:

        .. code-block:: python

            samplerate = 0.5
            cursor = np.array([[0,0], [1, 2], [2, 3], [3, 4], [4, 5], [5, 6]])
            eye = np.array([[1, 0], [1, 2], [1, 2], [4, 5], [4, 5], [6, 6]])
            targets = [
                np.array([[np.nan, np.nan], 
                        [5, 5], 
                        [np.nan, np.nan], 
                        [np.nan, np.nan], 
                        [5, 5], 
                        [np.nan, np.nan]]),
                np.array([[np.nan, np.nan], 
                        [np.nan, np.nan], 
                        [np.nan, np.nan], 
                        [-5, 5], 
                        [-5, 5], 
                        [-5, 5]])
            ]
            
            target_radius = 2.5
            target_colors = ['orange'] * len(targets)
            cursor_radius = 0.5
            bounds = [-10, 10, -10, 10]
            
            ani = animate_behavior(targets, cursor, eye, samplerate, bounds, target_radius, target_colors, cursor_radius, 
                            cursor_color='blue', eye_radius=0.25, eye_color='purple')
            

        .. raw:: html

            <video controls src="_static/test_anim_behavior.mp4"></video>
    '''

    fig, ax = plt.subplots(1, 1)

    # Use the animate_circles helper function 
    n_targets = len(targets)
    circle_radii = ([target_radius] * n_targets) + [cursor_radius, eye_radius]
    circle_colors = target_colors + [cursor_color, eye_color]
    circle_ts = targets + [cursor, eye]
    func = get_animate_circles_func(samplerate, bounds, circle_radii, circle_colors, 
                                    *circle_ts, history=history, ax=ax)      
    
    # Return the FuncAnimation object
    nframes = np.min([len(t) for t in circle_ts])
    ani = FuncAnimation(fig, func, frames=nframes, interval=1000./samplerate)  
    return ani


# controllers.py
# 
# Code relating to spectral analysis of feedforward & feedback controllers, based on the paper:
#
# Yamagami M, Peterson LN, Howell D, Roth E, Burden SA. Effect of Handedness on Learned Controllers 
# and Sensorimotor Noise During Trajectory-Tracking. IEEE Trans Cybern. 2023 Apr;53(4):2039-2050. 
# doi: 10.1109/TCYB.2021.3110187. Epub 2023 Mar 16. https://pubmed.ncbi.nlm.nih.gov/34587106/

import numpy as np
import math as m
from .base import calc_freq_domain_values

def get_machine_dynamics(freqs, order, exp_freqs=None):
    '''
    Returns the machine dynamics at each experimental frequency, given the system order. 

    Args:
        freqs (nt/2,): array of non-negative frequencies (essentially the x axis of a spectrogram)
        order (int): the order of the system (e.g. 0 for position control, 1 for velocity control, 2 for acceleration control)
        exp_freqs ((nfreq,), optional): list or array of frequencies used to generate experimental signals (reference and/or disturbance).
            If given, only the machine dynamics at these frequencies will be returned. Otherwise, by default, the machine dynamics at all
            frequencies will be returned.

    Returns:
        M ((nt/2,) or (nfreq,)): array of machine dynamics at each frequency

    '''
    s = 1.j*2*m.pi*freqs

    if order == 0:
        M = 1./np.ones((len(s),))
    elif order == 1:
        M = 1./(s)
    elif order == 2:
        M = 1./(s**2+s)
    else:
        raise ValueError(f"Order {order} not recognized!")

    if exp_freqs is not None:
        exp_idx = np.isin(freqs, exp_freqs)
        M = M[exp_idx]

    return M


def calc_transfer_function(input, output, samplerate, exp_freqs=None):
    '''
    Computes the frequency-domain transformation between time-domain input and ouput signals,
    at each frequency. Assumes output signals are produced by an LTI system.
    For math details, see Yamagami et al. IEEE Transactions on Cybernetics (2021).
    
    Args:
        input (nt, nch): time-domain input signal
        output (nt, nch): time-domain output signal
        samplerate (float): sampling rate of the data
        exp_freqs ((nfreq,), optional): list or array of frequencies used to generate experimental signals (reference and/or disturbance).
            If given, only the transformation at these frequencies will be returned. Otherwise, by default, the transformation at all
            frequencies will be returned.           

    Returns:
        tuple: Tuple containing:
            | **freqs ((nt/2,) or (nfreq,)):** array of frequencies (essentially the x axis of a spectrogram) 
            | **transfer_func ((nt/2, nch) or (nfreq, nch)):** array of complex numbers corresponding to the transformation at each of the above frequencies

    '''    
    assert len(input) == len(output), "Mismatched signal lengths"

    # Compute FFT of time-domain signals
    freqs, freq_input = calc_freq_domain_values(input, samplerate)
    _, freq_output = calc_freq_domain_values(output, samplerate)

    # Divide freq-domain output by freq-domain input at each frequency to find transformation
    transfer_func = np.divide(freq_output, freq_input)

    # Find the transformation at specific frequencies of interest
    if exp_freqs is not None:
        exp_freq_idx = np.isin(freqs.round(4), exp_freqs)
        freqs = freqs[exp_freq_idx]
        transfer_func = transfer_func[exp_freq_idx]

    return freqs.round(4), transfer_func


def pair_trials_by_frequency(ref_freqs, dis_freqs, max_trial_distance=1, limit_pairs_per_trial=True, max_pairs_per_trial=2):
    '''
    Finds pairs of trials with complementary frequency content (e.g. one trial's task signals consist of a subset
    of the experimental frequencies, and the other trial's task signals consist of the remaining experimental frequencies). 
    Assumes that reference and disturbance signals are made up of different frequencies within the same trial.

    Args:
        ref_freqs (ntrial,): list or array of frequencies used to generate the reference signal for each trial
        dis_freqs (ntrial,): list or array of frequencies used to generate the disturbance signal for each trial
        max_trial_distance (int, optional): maximum number of trials allowed between two trials identified as a pair
        limit_pairs_per_trial (bool, optional): whether to limit the number of pairs that any one trial can be included in
        max_pairs_per_trial (int, optional): maximum number of pairs that any one trial can be included in

    Returns:
        trial_pairs (npair, 2): array of trial indices corresponding to pairs of trials with complementary frequency content

    '''
    assert len(ref_freqs) == len(dis_freqs), "Mismatched number of trials"

    trial_pairs = []
    counts = np.zeros(len(ref_freqs),) # number of pairs each trial is included in

    for i in range(len(ref_freqs)):
        if limit_pairs_per_trial and counts[i] >= max_pairs_per_trial:
            continue
        curr_r = np.array(ref_freqs[i]) # convert to array in case in list format
        curr_d = np.array(dis_freqs[i])

        # Find next trial with complementary frequency content
        for j in range(i+1,len(ref_freqs)):
            if j-i > max_trial_distance:
                break
            if limit_pairs_per_trial and counts[j] >= max_pairs_per_trial:
                continue
            next_r = np.array(ref_freqs[j])
            next_d = np.array(dis_freqs[j])

            if (next_r != curr_r).all() and (next_d != curr_d).all():
                trial_pairs.append([i,j])
                counts[i] += 1
                counts[j] += 1
                    
    return np.array(trial_pairs)


def calc_F_B_controllers(usr, ref, dis, exp_freqs, ref_freqs, dis_freqs, samplerate, system_order, trial_pairs=None):
    '''
    Estimates a user's frequency-domain feedforward and feedback controllers at each experimental frequency in a 
    reference-tracking, disturbance-rejection task. Assumes the user responds to reference & disturbance signals like an LTI system.
    For math details, see Yamagami et al. IEEE Transactions on Cybernetics (2021).
    
    Args:
        usr (ntrial, nt): array of time-domain user response signals from each trial
        ref (ntrial, nt): array of time-domain reference signals from each trial 
        dis (ntrial, nt): array of time-domain disturbance signals from each trial
        exp_freqs (nfreq,): list or array of frequencies used to generate experimental signals (reference and/or disturbance)   
        ref_freqs (ntrial,): list or array of frequencies used to generate the reference signal for each trial
        dis_freqs (ntrial,): list or array of frequencies used to generate the disturbance signal for each trial
        samplerate (float): sampling rate of the data
        system_order (int): the order of the system (e.g. 0 for position control, 1 for velocity control, 2 for acceleration control)
        trial_pairs ((npair, 2), optional): list or array of trial indices corresponding to pairs of trials with complementary frequency content.
            If given, controllers will be computed over these trial pairs. Otherwise, by default, controllers will be computed over trials paired 
            up using the default parameters of :func:`~aopy.analysis.controllers.pair_trials_by_frequency`.   
        
    Returns:
        tuple: tuple containing:
            | **F (npair, nfreq):** array of complex numbers corresponding to the user's feedforward controller at each experimental frequency 
            | **B (npair, nfreq):** array of complex numbers corresponding to the user's feedback controller at each experimental frequency
            | **Tur (npair, nfreq):** array of complex numbers corresponding to the ref-->user transformation at each experimental frequency
            | **Tud (npair, nfreq):** array of complex numbers corresponding to the dis-->user transformation at each experimental frequency
            | **M (nfreq,):** array of machine dynamics at each experimental frequency

    '''
    assert len(usr) == len(ref) == len(dis), "Mismatched number of trials"

    # Calculate transfer functions from ref-->user and dis-->user
    trial_Tur = np.array([np.squeeze(calc_transfer_function(ref[i], usr[i], samplerate, exp_freqs)[1]) for i in range(len(usr))])
    trial_Tud = np.array([np.squeeze(calc_transfer_function(dis[i], usr[i], samplerate, exp_freqs)[1]) for i in range(len(usr))])  

    # Get machine dynamics, given system order
    freqs = calc_freq_domain_values(usr[0], samplerate)[0]
    M = get_machine_dynamics(freqs.round(4), system_order, exp_freqs)

    # Pair up trials by frequency content
    if trial_pairs is None:
        trial_pairs = pair_trials_by_frequency(ref_freqs, dis_freqs)

    # Find transfer functions and controllers at the experimental frequencies for each trial pair
    Tur = np.zeros((len(trial_pairs),len(exp_freqs)), dtype=complex)*np.nan
    Tud = np.zeros((len(trial_pairs),len(exp_freqs)), dtype=complex)*np.nan
    B = np.zeros((len(trial_pairs),len(exp_freqs)), dtype=complex)*np.nan
    F = np.zeros((len(trial_pairs),len(exp_freqs)), dtype=complex)*np.nan
    
    for pair_id, (trial_a, trial_b) in enumerate(trial_pairs):
        ref_ind_a = np.isin(exp_freqs, ref_freqs[trial_a])
        dis_ind_a = np.isin(exp_freqs, dis_freqs[trial_a])
        ref_ind_b = np.isin(exp_freqs, ref_freqs[trial_b])
        dis_ind_b = np.isin(exp_freqs, dis_freqs[trial_b])

        Tur[pair_id, ref_ind_a] = trial_Tur[trial_a, ref_ind_a]
        Tur[pair_id, ref_ind_b] = trial_Tur[trial_b, ref_ind_b]
        Tud[pair_id, dis_ind_a] = trial_Tud[trial_a, dis_ind_a]
        Tud[pair_id, dis_ind_b] = trial_Tud[trial_b, dis_ind_b]

        # Compute feedback (B) and feedforward (F) controllers - see eq. 6a and 6b in Yamagami et al., 2023
        # B = -Tud/(M*(1 + Tud))
        # F = (1 + B*M)*(Tur) - B
        B[pair_id,:] = np.divide( -Tud[pair_id], np.multiply( M, np.ones(Tud[pair_id].shape,dtype=complex)+Tud[pair_id] ) )
        F[pair_id,:] = np.multiply( Tur[pair_id], (1+0j)+np.multiply( B[pair_id,:], M ) ) - B[pair_id,:]

    return F, B, Tur, Tud, M


def calc_crossover_freq(freqs, M, B):
    '''
    Estimate crossover frequency, the maximum frequency at which a user can properly 
    track the target. Above crossover, reference-tracking and disturbance-rejection 
    performance degrades. 
    Based on the McRuer crossover model, crossover frequency is the experimental frequency
    at which the gain of the user's open-loop transfer function equals 1 (see McRuer & Jex, 1967).
    Below crossover, the gain is >1. Above crossover, the gain is <1.
    
    Args:
        freqs ((nfreqs,) array): experimental frequencies used to generate reference and disturbance trajectories
        M ((nfreqs,) array): machine dynamics corresponding to the system order (0th order, 1st order, etc.)
        B ((ntrials, nfreqs) complex array): user's feedback controller for each trial (including more trials produces more accurate estimate)
    
    Returns:
        tuple: tuple containing:
        | **xover_freq (float):** estimated crossover frequency
        | **L ((nfreqs,) array):** gain of user's open-loop transfer function
    
    '''
    # estimate the user's open-loop transfer function
    L = np.zeros((len(freqs),))
    for i in range(len(freqs)):
        # |L(w)| = |M(w)*B(w)|
        L[i] = np.mean(abs(M[i]*B[:,i]))
    above_xover_ind = np.where(L<1)[0]
    if len(above_xover_ind) < 1: 
        # crossover occurs at a freq higher than tested
        xover_freq = np.nan
    else:
        # crossover is the lowest freq where L<1
        xover_freq = freqs[above_xover_ind[0]]
    return xover_freq, L


import numpy as np

def get_saccade_pos(eye_pos, onset_times, duration, samplerate):
    '''
    Returns the coordinates of the start and end of each given saccade
    
    Args:
        eye_pos (nt,nch): eye position data
        onset_times (nsaccade): saccade onset times (in seconds) generated in a trial segment.
        duration (nsaccade): saccade duration (in seconds) of saccades generated in a trial segment
        samplerate (float): sampling rate
    
    Returns:
        tuple: tuple containing:
            | **onset_pos (nsaccade,nch):** eye positions when saccades start in a given trial
            | **offset_pos (nsaccade,nch):** eye positions when saccades end in a given trial 
    '''    

    # Convert onset times and durations to sample indices
    onset_indices = (onset_times * samplerate).astype(int)
    offset_indices = ((onset_times + duration) * samplerate).astype(int)
    
    # Get eye positions for onset and offset indices
    onset_pos = eye_pos[onset_indices, :]
    offset_pos = eye_pos[offset_indices, :]
    
    return onset_pos, offset_pos
        
def get_saccade_target_index(onset_pos, offset_pos, target_pos, target_radius):
    '''
    Determines a target index a subject looked at during a saccade. When the subject 
    looked at irrelevant areas, the target index is -1.
    
    Args:
        onset_pos (nsaccade, 2): eye positions when saccades start in a given trial
        offset_pos (nsaccade, 2): eye positions when saccades end in a given trial
        target_pos (ntarget, 2): target positions, e.g. center and peripheral targets
        target_radius (float): radius around the target center to search
    
    Returns:
        tuple: tuple containing:
            | **onset_target (nsaccade):** target index at saccade start
            | **offset_target (nsaccade):** target index at saccade end    
    '''
    
    # Calculate distances between saccade positions and target positions
    onset_dists = np.linalg.norm(onset_pos[:, np.newaxis, :] - target_pos, axis=2)
    offset_dists = np.linalg.norm(offset_pos[:, np.newaxis, :] - target_pos, axis=2)
    
    # Determine if the saccades are directed towards any target
    onset_label = onset_dists <= target_radius
    offset_label = offset_dists <= target_radius
    
    # Get target numbers for onset and offset
    onset_target = np.argmax(onset_label, axis=1)
    offset_target = np.argmax(offset_label, axis=1)
    
    # Set target number to -1 for saccades not directed at any target
    onset_target[~np.any(onset_label, axis=1)] = -1
    offset_target[~np.any(offset_label, axis=1)] = -1
    
    return onset_target, offset_target

def get_saccade_event(onset_times, duration, event_times, event_codes):
    '''
    Returns event codes corresponding to times in which the start and end of 
    each given saccade falls. 
    
    Args:
        onset_times (nsaccade): saccade onset times (in seconds)
        duration (nsaccade): saccade duration (in seconds) of saccades in the onset_times list
        event_times (nevent): list of event times (in seconds)
        event_codes (nevent): event codes corresponding to the given event_times
    
    Returns:
        tuple: tuple containing:
            | **onset_event (nsaccade):** event in which each saccade starts
            | **offset_event (nsaccade):** event in which each saccade ends
    '''
    
    # Calculate saccade offset times
    offset_times = onset_times + duration
    
    # Each index is the event preceding the saccade
    onset_indices = np.searchsorted(event_times, onset_times, side='right') - 1
    offset_indices = np.searchsorted(event_times, offset_times, side='right') - 1
    
    # Get event codes using the indices
    onset_event = event_codes[onset_indices]
    offset_event = event_codes[offset_indices]
    
    return onset_event, offset_event


import platform

def get_memory_available_gb():
    '''
    Get the available system memory in gigabytes. Only works on linux platforms.

    Note:
        The results of this function are equivalent to the terminal commands:
        * "grep MemAvailable /proc/meminfo" -> available memory
        * "grep MemTotal /proc/meminfo" -> total memory

    Returns:
        int: number of gigabytes of available system memory
    '''
    if platform.system() != "Linux":
        print('Only works on linux!')
        return
    with open('/proc/meminfo', 'r') as mem:
        free_memory = 0
        for i in mem:
            sline = i.split()
            if str(sline[0]) == 'MemAvailable:':
                free_memory = int(sline[1])
                break
    return int(free_memory/1e6) # convert KB to GB

def set_memory_limit_gb(size_gb):
    '''
    Set a memory resource limit in gigabytes. Only works on linux platforms.

    Note: 
        This function sets a soft limit, not a hard limit. The soft limit is a value upon 
        which the operating system will restrict memory usage by the process (python, 
        in this case). A true upper bound on the memory values can be defined by the hard 
        limit. However, although the hard limit can be lowered, it can never be raised by 
        user processes (even if the process lowered itself) and is controlled by a 
        system-wide parameter set by the system administrator. Nevertheless, the soft
        limit should serve to raise a `MemoryError` whenever python exceeds the setting.

    Args:
        size_gb (int): upper limit of memory that will be made available to python in gigabytes
    '''
    if platform.system() != "Linux":
        print('Only works on linux!')
        return
    import resource
    maxsize = int(1e9*size_gb)
    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    resource.setrlimit(resource.RLIMIT_AS, (maxsize, hard))

def get_memory_limit_gb():
    '''
    Get the memory resource limit in gigabytes. Only works on linux platforms.

    Returns:
        int or None: upper limit of memory available to python in gigabytes
    '''
    if platform.system() != "Linux":
        print('Only works on linux!')
        return
    import resource
    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    if soft == resource.RLIM_INFINITY:
        return
    return int(soft/1e9)
    
def release_memory_limit():
    '''
    Unset any memory resource limit that may have been applied. Only works on 
    linux platforms.
    '''
    if platform.system() != "Linux":
        print('Only works on linux!')
        return
    import resource
    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    resource.setrlimit(resource.RLIMIT_AS, (resource.RLIM_INFINITY, hard))



from ..data.bmi3d import load_ecube_data_chunked
from ..precondition.eye import filter_eye
from .. import precondition
from .. import data as aodata
from .. import utils
import numpy as np
import os

def parse_oculomatic(data_dir, files, samplerate=1000, max_memory_gb=1.0, debug=True, **filter_kwargs):
    """
    Loads eye data from ecube and hdf data. 

    .. image:: _images/proc_oculomatic_downsample.png
    .. image:: _images/proc_oculomatic_mask.png
    
    Data includes:
        data (nt, nch): eye data in volts
        mask (nt, nch): boolean mask of when the eyes are closed
        
    Metadata includes:
        channels (nch): analog channels on which eye data was recorded
        labels (nch): string labels for each eye channel
        samplerate (float): sampling rate of the eye data
        units (str): measurement unit of eye data

    Args:
        data_dir (str): folder containing the data you want to load
        files (dict): a dictionary that has 'ecube' as the key
        samplerate (float, optional): sampling rate to output in Hz. Default 1000. 
        max_memory_gb (float, optional): max memory used to load binary data at one time
        debug (bool, optional): prints debug information
        filter_kwargs (kwargs, optional): optional keyword arguments to send to `filter_eye()`
        
    Returns:
        tuple: tuple contatining:
            | **eye_data (nt, neyech):** voltage per eye channel (normally [left eye x, left eye y, right eye x, right eye y])
            | **eye_metadata (dict):** metadata associated with the eye data, including the above labels
    """
    
    eye_metadata = dict()
    
    if 'hdf' in files:
        bmi3d_events, bmi3d_event_metadata = aodata.load_bmi3d_hdf_table(data_dir, files['hdf'], 'sync_events')

        # get eye channels 
        if 'left_eye_ach' in bmi3d_event_metadata and 'right_eye_ach' in bmi3d_event_metadata:
            eye_channels = bmi3d_event_metadata['left_eye_ach'] + bmi3d_event_metadata['right_eye_ach']
            if debug: print(f'use bmi3d supplied eye channel definition {eye_channels}')
        else:
            eye_channels = [9, 8, 10, 11]
            if debug: print(f'eye channel definitions do not exist, use eye channels {eye_channels} ')
    else:
        # from https://github.com/aolabNeuro/analyze/issues/225
        eye_channels = [10, 11, 8, 9]
        if debug: print(f'No metadata from BMI3D, assuming eye channels {eye_channels} ')
        
    eye_metadata['source'] = 'oculomatic voltage output'
    eye_metadata['n_channels'] = len(eye_channels)
    eye_metadata['channels'] = eye_channels
    eye_metadata['labels']  = ['left_eye_x', 'left_eye_y', 'right_eye_x', 'right_eye_y']
    
    # Create an empty array for the downsampled data
    data_path = os.path.join(data_dir, files['ecube'])
    analog_metadata = aodata.load_ecube_metadata(data_path, 'AnalogPanel')
    dtype = 'float'
    chunksize = int(max_memory_gb * 1e9 / np.dtype(dtype).itemsize / eye_metadata['n_channels'])
    downsample_factor = int(analog_metadata['samplerate']/samplerate)
    n_samples = int(np.ceil(analog_metadata['n_samples']/downsample_factor))
    n_channels = int(eye_metadata['n_channels'])
    downsample_data = np.zeros((n_samples, n_channels), dtype=dtype)
    downsample_mask = np.zeros((n_samples, n_channels), dtype='bool')

    # Filter broadband data into LFP directly into the hdf file
    n_samples = 0
    for analog_chunk in load_ecube_data_chunked(data_path, 'AnalogPanel', channels=eye_channels, chunksize=chunksize):
        mask_chunk = detect_noise(analog_chunk, analog_metadata['samplerate'], min_step=3, step_thr=3)
        # mask_chunk = precondition.downsample(mask_chunk, analog_metadata['samplerate'], samplerate)
        # mask_chunk = mask_chunk > 0.5 # downsample takes a mean over boolean, this makes it back into a boolean
        mask_chunk = mask_chunk[::downsample_factor,:]
        analog_chunk, _ = filter_eye(analog_chunk, analog_metadata['samplerate'], downsamplerate=samplerate, **filter_kwargs)
        chunk_len = analog_chunk.shape[0]
        downsample_data[n_samples:n_samples+chunk_len,:] = analog_chunk
        downsample_mask[n_samples:n_samples+chunk_len,:] = mask_chunk
        n_samples += chunk_len

    eye_metadata['source'] = data_path
    eye_metadata['raw_samplerate'] = analog_metadata['samplerate']
    eye_metadata['samplerate'] = samplerate
    eye_metadata['n_samples'] = downsample_data.shape[0]
    eye_metadata['taper_len'] = 0.05
    eye_metadata['lowpass_freq'] = 30
    eye_metadata['pad_t'] = 1.0
    eye_metadata.update(filter_kwargs)
    
    #scale eye data from bits to volts
    if 'voltsperbit' in analog_metadata:
        analog_voltsperbit = analog_metadata['voltsperbit']
    else:
        analog_voltsperbit = 3.0517578125e-4
    eye_metadata['units'] = 'volts'
        
    eye_data = {
        'data': downsample_data * analog_voltsperbit,
        'mask': downsample_mask
    }
    return eye_data, eye_metadata

def detect_noise(eye_data, samplerate, min_step=None, step_thr=3, t_closed_min=0.1):
    '''
    Detect noise in oculomatic eye data. Searches for repeated data which indicates
    that the subject's eyes were closed and returns a mask for each eye. Uses the
    minimum step size in the data to estimate when the voltage has "changed" versus
    when it is fluctuating because of measurement noise.
    
    Args:
        eye_data ((nt,nch) array): unfiltered raw or calibrated eye position data
        samplerate (float): sampling rate of the eye data
        min_step (float, optional): minimum step size in the data. If None, it will be 
            calculated automatically. Default None.
        step_thr (float, optional): multiple of the minimum step size to use as a threshold for 
            detecting changing values in the eye data
        t_closed_min (float, optional): number of seconds the data must remain unchanged to be
            included in the eye closed mask
        
    Returns:
        (nt, nch) eye_closed_mask: boolean mask, True for each eye when it was probably closed.
    '''
    eye_closed_mask = np.zeros(eye_data.shape, dtype='bool')

    # Detect when the eyes are closed -- oculomatic specific
    for eye_idx in range(eye_data.shape[1]):
        
        # Find where value changes from one sample to next
        diff = abs(np.diff(eye_data[:,eye_idx]))
        if min_step is None:
            step_size = step_thr*np.min(diff[diff > 0])
        else:
            step_size = step_thr*min_step
        repetitions, repetitions_idx = utils.count_repetitions(eye_data[:,eye_idx], step_size)
        # Mask anything that is longer than a predetermined length
        suspicious = repetitions > int(samplerate*t_closed_min)
        suspicious_idx = repetitions_idx[suspicious]
        durations = repetitions[suspicious]
        for idx in range(len(suspicious_idx)):
            idx_start = suspicious_idx[idx]
            idx_end = idx_start + durations[idx]
            eye_closed_mask[idx_start:idx_end,eye_idx] = 1
            
    return eye_closed_mask



# tuning.py
#
# Code related to tuning analysis, e.g. modulation depth, specificity, curve fitting, etc.

import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import f_oneway


'''
Curve fitting
'''
# These functions are for curve fitting and getting modulation depth and preferred direction from firing rates
def curve_fitting_func(target_theta, b1, b2, b3):
    '''

    Args:
        target_theta (float): center out task target direction index [degrees]
        b1, b2, b3 (float): parameters used for curve fitting

    .. math::
    
        b1 * cos(\\theta) + b2 * sin(\\theta) + b3

    Returns:
        float: Evaluation of the fitting function for a given target.

    '''
    return b1 * np.cos(np.deg2rad(target_theta)) + b2 * np.sin(np.deg2rad(target_theta)) + b3


def get_modulation_depth(b1, b2):
    '''
    Calculates modulation depth from curve fitting parameters as follows:
    
    .. math::
    
        \\sqrt{b_1^2+b_2^2}

    Returns:
        float: Modulation depth (amplitude) of the curve fit
    '''
    return np.sqrt((b1 ** 2) + (b2 ** 2))


def get_preferred_direction(b1, b2):
    '''
    Calculates preferred direction from curve fitting parameters as follows:
    
    .. math:: 
        
        arctan(\\frac{b_1^2}{b_2^2})
    
    Returns:
        float: Preferred direction of the curve fit in radians
    '''
    b1sign = np.sign(b1)
    b2sign = np.sign(b2)
    temp_pd = np.arctan2(b2sign*b2**2, b1sign*b1**2)
    if temp_pd < 0:
      pd = (2*np.pi)+temp_pd
    else:
      pd = temp_pd
    return pd


def get_mean_fr_per_condition(data, condition_labels, return_significance=False):
    '''
    This function computes the average activity for each feature and trial. 

    Args:
        data (ntime, nch, ntrials): Trial aligned neural data
        condition_labels (ntrials): condition label for each trial
        return_significance (bool): Uses the one-way ANOVA test to compute a p-value for each channel/unit

    Returns:
        tuple: Tuple containing:
            | **means_d: (nch, nconditions)** = mean firing rate per neuron per target direction
            | **stds_d: (nch, nconditions)** standard deviation from mean firing rate per neuron
            | **pvalue: (nch)** significance of modulation
    '''
    means_d = []
    stds_d = []
    unique_condition_labels = np.unique(condition_labels)
    nconditions = len(unique_condition_labels)

    [means_d.append(np.mean(data[:,:,condition_labels==unique_condition_labels[icond]], axis=(0,2))) for icond in range(nconditions)]
    [stds_d.append(np.std(data[:,:,condition_labels==unique_condition_labels[icond]], axis=(0,2))) for icond in range(nconditions)]    

    if return_significance:
        cond_means = [] 
        [cond_means.append(np.mean(data[:,:,condition_labels==unique_condition_labels[icond]], axis=0)) for icond in range(nconditions)]
        _, pvalue = f_oneway(*cond_means, axis=1)
        return np.array(means_d).T, np.array(stds_d).T, pvalue
    else:
        return np.array(means_d).T, np.array(stds_d).T
    

def run_tuningcurve_fit(mean_fr, targets, fit_with_nans=False, min_data_pts=3):
    '''
    This function calculates the tuning parameters from center out task neural data.
    It fits a sinusoidal tuning curve to the mean firing rates for each unit.
    Uses approach outlined in Orsborn 2014/Georgopolous 1986.
    Note: To orient PDs with the quadrant of best fit, this function samples the target location
    with high resolution between 0-360 degrees.

    Args:
        mean_fr (nunits, ntargets): The average firing rate for each unit for each target.
        targets (ntargets): Targets locations to fit to [Degrees]. Corresponds to order of targets in 'mean_fr' (Xaxis in the fit). Targets should be monotonically increasing.
        fit_with_nans (bool): Optional. Control if the curve fitting should be performed for a unit in the presence of nans. If true curve fitting will be run on non-nan values but will return nan is less than 3 non-nan values. If false, any unit that contains a nan will have the corresponding outputs set to nan.
        min_data_pts (int): Optional. 

    Returns:
        tuple: Tuple containing:
            | **fit_params (3, nunits):** Curve fitting parameters for each unit
            | **modulation depth (ntargets, nunits):** Modulation depth of each unit
            | **preferred direction (ntargets, nunits):** preferred direction of each unit [rad]
    '''
    nunits = np.shape(mean_fr)[0]
    ntargets = len(targets)

    fit_params = np.empty((nunits,3))*np.nan
    md = np.empty((nunits))*np.nan
    pd = np.empty((nunits))*np.nan

    for iunit in range(nunits):
        # If there is a nan in the values of interest skip curve fitting, otherwise fit
        if ~np.isnan(mean_fr[iunit,:]).any():
            params, _ = curve_fit(curve_fitting_func, targets, mean_fr[iunit,:])
            fit_params[iunit,:] = params

            md[iunit] = get_modulation_depth(params[0], params[1])
            pd[iunit] = get_preferred_direction(params[0], params[1])

        # If this doesn't work, check if fit_with_nans is true. It it is remove nans and fit
        elif fit_with_nans:
            nonnanidx = ~np.isnan(mean_fr[iunit,:])
            if np.sum(nonnanidx) >= min_data_pts: # If there are enough data points run curve fitting, else return nan
                params, _ = curve_fit(curve_fitting_func, targets[nonnanidx], mean_fr[iunit,nonnanidx])
                fit_params[iunit,:] = params

                md[iunit] = get_modulation_depth(params[0], params[1])
                pd[iunit] = get_preferred_direction(params[0], params[1])
            else:
                md[iunit] = np.nan
                pd[iunit] = np.nan

    return fit_params, md, pd



import numpy as np
import matplotlib.pyplot as plt

from .. import postproc

def plot_eye_calibration_result(eye_calibration_data, cursor_calibration_data, coefficients, correlation_coeff,
                                eye_labels=['LE x', 'LE y', 'RE x', 'RE y'], bounds=[-10, 10, -10, 10]):
    '''
    Plot the results of eye calibration.

    Args:
        eye_calibration_data (nt, nch): The eye calibration data.
        cursor_calibration_data (nt, 2): The cursor calibration data.
        coefficients (nch): The coefficients for eye calibration.
        correlation_coeff (float): The correlation coefficient.
        eye_labels (list, optional): The labels for each eye. Defaults to ['LE x', 'LE y', 'RE x', 'RE y'].
        bounds (list, optional): The bounds for the plot. Defaults to [-10, 10, -10, 10].

    Examples:

        Beignet center-out task (5974) from 2022-07-01:
        .. image:: _images/eye_calibration.png
    '''
    
    estimated_calibration_pos = postproc.get_calibrated_eye_data(eye_calibration_data, coefficients)

    nch = eye_calibration_data.shape[1]
    fig = plt.figure(figsize=(nch*3,6))
    gs = fig.add_gridspec(2,nch)
    for i in range(nch):
        ax = fig.add_subplot(gs[0,i])
        eye_label = eye_labels[i]
        cursor_label = ['cursor x', 'cursor y'][i % 2]
        
        ax.set_title(f'{eye_label} vs. {cursor_label}')
        ax.scatter(cursor_calibration_data[:,i], eye_calibration_data[:,i], color='b')
        ax.plot(estimated_calibration_pos[:,i], eye_calibration_data[:,i], 'k')
        ax.text(0.6, 0.9, f'$R^2: {correlation_coeff[i]**2:.4f}$', transform=ax.transAxes)
        ax.text(0.05, 0.05, f' \n $y = {coefficients[i,0]:.2f} * x - {coefficients[i,1]:.2f}$ ', transform=ax.transAxes)
        ax.set_ylabel("Eye raw pos.")
        ax.set_xlabel('Cursor pos. (cm)')

    ax = fig.add_subplot(gs[1,:nch//2])
    ax.scatter(cursor_calibration_data[:,0], cursor_calibration_data[:,1], color='g', label='cursor')
    ax.scatter(estimated_calibration_pos[:,0], estimated_calibration_pos[:,1], color='m', label='eye')
    ax.set_title('LE calibration')
    ax.set_xlim(bounds[0], bounds[1])
    ax.set_ylim(bounds[2], bounds[3])
    ax.set_aspect('equal')
    handles, labels = ax.get_legend_handles_labels()
    unique = [h for i, h in enumerate(handles) if h not in handles[:i]]
    plt.legend(unique, ['cursor', 'eye'], loc='upper left', bbox_to_anchor=(1,1))
    
    ax = fig.add_subplot(gs[1,nch//2:])
    ax.scatter(cursor_calibration_data[:,0], cursor_calibration_data[:,1], color='g', label='cursor')
    ax.scatter(estimated_calibration_pos[:,2], estimated_calibration_pos[:,3], color='m', label='eye')
    ax.set_title('RE calibration')
    ax.set_xlim(bounds[0], bounds[1])
    ax.set_ylim(bounds[2], bounds[3])
    ax.set_aspect('equal')
    handles, labels = ax.get_legend_handles_labels()
    unique = [h for i, h in enumerate(handles) if h not in handles[:i]]
    plt.legend(unique, ['cursor', 'eye'], loc='upper left', bbox_to_anchor=(1,1))

    plt.tight_layout()


# visualization.py
# Code for general neural data plotting (raster plots, multi-channel field potential plots, psth, etc.)
import string
import warnings
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib import cm
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection

from scipy.interpolate import griddata
from scipy.interpolate.interpnd import _ndim_coords_from_arrays
from scipy.spatial import cKDTree
from scipy import signal
from scipy.stats import zscore
import numpy as np
import os
from PIL import Image
import copy
import pandas as pd
from tqdm import tqdm
import pandas as pd
from datetime import timedelta

from .. import precondition
from .. import analysis
from .. import data as aodata
from .. import utils
from .. import preproc

def plot_mean_fr_per_target_direction(means_d, neuron_id, ax, color, this_alpha, this_label):
    '''
    generate a plot of mean firing rate per target direction

    '''
    sns.set_context('talk')
    ax.plot(np.array(means_d)[:, neuron_id], c=color, alpha=this_alpha, label=this_label)

    ax.legend()
    ax.set_xlabel("Target", fontsize=16)
    ax.set_ylabel("Spike Rate (Hz)", fontsize=16)
    plt.tight_layout()


def savefig(base_dir, filename, **kwargs):
    '''
    Wrapper around matplotlib savefig with some default options

    Args:
        base_dir (str): where to put the figure
        filename (str): what to name the figure
        **kwargs (optional): arguments to pass to plt.savefig()
    '''
    if '.' not in filename:
        filename += '.png'
    fname = os.path.join(base_dir, filename)
    if 'dpi' not in kwargs:
        kwargs['dpi'] = 300.
    if 'edgecolor' not in kwargs:
        kwargs['edgecolor'] = 'none'
    if 'transparent' not in kwargs:
        kwargs['transparent'] = True
    if kwargs['transparent'] and 'facecolor' not in kwargs:
        kwargs['facecolor'] = 'none'
    plt.savefig(fname, **kwargs)

def subplots_with_labels(n_rows, n_cols, return_labeled_axes=False, 
                         rel_label_x=-0.25, rel_label_y=1.1,
                         label_font_size=11, constrained_layout=False, 
                         **kwargs):
    '''
    Create a figure with subplots labeled with letters. Augments plt.subplots().

    Examples:

        Generate a figure with 2 rows and 2 columns of subplots, labeled A, B, C, D

        .. code-block:: python
            
            fig, axes = subplots_with_labels(2, 2, constrained_layout=True)

        .. image:: _images/labeled_subplots.png

    Args:
        n_rows (int): Number of rows of subplots.
        n_cols (int): Number of columns of subplots.
        return_labeled_axes (bool, optional): Whether to return the labeled axes. Default False.
        rel_label_x (float, optional): The relative x position of the subplot label. Default -0.25.
        rel_label_y (float, optional): The relative y position of the subplot label. Default 1.1
        label_font_size (int, optional): The font size of the subplot label. Default 11.
        constrained_layout (bool, optional): Whether to use constrained layout. Default is False.
        **kwargs: Additional keyword arguments to pass to plt.subplot_mosaic.

    Returns:
        fig (Figure): The created figure.
        axes (np.ndarray): The created axes.
        labels_axes (dict, optional): The labeled axes if return_labeled_axes is True.
    '''
    # if more than 26 subplots, raise an error
    if n_rows * n_cols > 26:
        raise ValueError("More than 26 subplots requested, running out of single letters to label them with!")

    # make a list of letters to use as labels
    alphabets = string.ascii_uppercase
    labels = alphabets[:n_rows * n_cols]

    # tabulate the labels into n_rows by n_cols array
    labels = np.array(list(labels)).reshape((n_rows, n_cols))

    # make a string where rows are separated by semicolons
    labels = ";".join(["".join(row) for row in labels])

    # make the figure and axes
    fig, labels_axes = plt.subplot_mosaic(labels, constrained_layout=constrained_layout, **kwargs)

    for n, (key, ax) in enumerate(labels_axes.items()):
        ax.text(rel_label_x, rel_label_y, key, transform=ax.transAxes, size=label_font_size)

    # just annotate the axes
    axes = list(labels_axes.values())
    axes = np.array(axes).reshape((n_rows, n_cols))

    if return_labeled_axes:
        return fig, axes, labels_axes
    else:
        return fig, axes

def place_subplots(fig, positions, width, height, **kwargs):
    '''
    Plotting utility to create subplots in arbitrary positions on a figure. Positions 
    are in inches from the bottom left corner of the figure.
    
    Args:
        fig (pyplot.Figure): figure to place the subplots on
        positions (npos, 2): list of (x, y) coordinates (in inches) where to center the subplots 
        width (float): width (in inches) of each subplot
        height (float): height (in inches) of each subplot
        kwargs (dict, optional): other keyword arguments to pass to fig.add_axes
        
    Returns:
        list: pyplot.Axes handles for each position

    Examples:

        .. code-block:: python

            fig = plt.figure(figsize=(4,6))
            positions = [[1, 2], [3, 4]]
            width = 1
            height = 1
            ax = place_subplots(fig, positions, width, height)
            ax[0].annotate('1', (0.5,0.5), fontsize=40)
            ax[1].annotate('2', (0.5,0.5), fontsize=40)

        .. image:: _images/place_subplots_1.png

        .. code-block:: python
   
            fig = plt.figure(figsize=(4,6))
            positions = [[1, 1.5], [3, 4.5]]
            width = 2
            height = 3
            ax = place_subplots(fig, positions, width, height)
            ax[0].annotate('1', (0.5,0.5), fontsize=40)
            ax[1].annotate('2', (0.5,0.5), fontsize=40)

        .. image:: _images/place_subplots_2.png
            
    '''
    # Normalize the positions to fit into the size of the figure
    fig_width, fig_height = fig.get_size_inches()
    positions = np.array(positions, dtype='float')
    positions[:,0] = positions[:,0] / fig_width
    positions[:,1] = positions[:,1] / fig_height
    width /= fig_width
    height /= fig_height

    # Place subplots
    ax = []
    for cx, cy in positions:
        left = cx - width/2
        bottom = cy - height/2
        ax.append(fig.add_axes([left, bottom, width, height], **kwargs))
    return ax

def place_Opto32_subplots(fig_size=5, subplot_size=0.75, offset=(0.,-0.25), **kwargs):
    '''
    Wrapper around place_subplots() for the Opto32 stimulation sites.

    Args:
        fig_size (float): width and height (in inches) of the figure
        subplot_size (float): width and height (in inches) of each subplot
        offset (tuple): x and y offset (in inches) from the bottom left corner of the figure
        kwargs (dict, optional): other keyword arguments to pass to fig.add_axes

    Returns:
        tuple: tuple containing:
        | **fig (pyplot.Figure):** figure where the subplots were placed
        | **ax (list):** pyplot.Axes handles for each stimulation site

    Examples:

        .. image:: _images/place_Opto32_subplots.png
    '''
    stim_pos, _, _ = aodata.load_chmap('Opto32')

    # Normalize the positions to the width and height of the figure
    stim_pos = (stim_pos - np.mean(stim_pos, axis=0)) / (np.max(stim_pos) - np.min(stim_pos)) * fig_size + fig_size/2

    # Place subplots
    fig = plt.figure(figsize=(fig_size,fig_size), **kwargs)
    ax = place_subplots(fig, stim_pos + np.array(offset), subplot_size, subplot_size)

    # Remove the axis labels
    for ax_ in ax:
        ax_.tick_params(
            which='both',
            bottom=False,
            left=False,  
            labelbottom=False,
            labelleft=False
        )
    return fig, ax

def plot_timeseries(data, samplerate, t0=0., ax=None, **kwargs):
    '''
    Plots data along time on the given axis. Default units are seconds and volts.

    Example:
        
        Plot 50 and 100 Hz sine wave.

        .. code-block:: python

            data = np.reshape(np.sin(np.pi*np.arange(1000)/10) + np.sin(2*np.pi*np.arange(1000)/10), (1000))
            samplerate = 1000
            plot_timeseries(data, samplerate)

        .. image:: _images/timeseries.png

    Args:
        data (nt, nch): timeseries data in volts, can also be a single channel vector
        samplerate (float): sampling rate of the data
        t0 (float, optional): time (in seconds) of the first sample. Default 0.
        ax (pyplot axis, optional): where to plot
        kwargs (dict, optional): optional keyword arguments to pass to plt.plot
    '''
    if np.ndim(data) < 2:
        data = np.expand_dims(data, 1)
    if ax is None:
        ax = plt.gca()

    time = np.arange(np.shape(data)[0]) / samplerate + t0
    for ch in range(np.shape(data)[1]):
        ax.plot(time, data[:, ch], **kwargs)
    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Voltage (V)')

def gradient_timeseries(data, samplerate, n_colors=100, color_palette='viridis', ax=None, **kwargs):
    '''
    Draw gradient lines of timeseries data. Default units are seconds and volts.

    Args:
        data (nt, nch): timeseries to plot, can be 1d or 2d.
        samplerate (float): sampling rate of the data
        n_colors (int, optional): number of colors in the gradient. Default 100.
        color_palette (str, optional): colormap to use for the gradient. Default 'viridis'.
        ax (plt.Axis, optional): axis to plot the targets on
        kwargs (dict): keyword arguments to pass to the LineCollection function (similar to plt.plot)

    Raises:
        ValueError: if the data has more than 2 dimensions

    Example:
        .. code-block:: python

            data = np.reshape(np.sin(np.pi*np.arange(1000)/100), (1000))
            samplerate = 1000
            gradient_timeseries(data, samplerate)

        .. image:: _images/timeseries_gradient.png
    '''           
    if data.ndim == 1:
        data = np.expand_dims(data, 1)
    elif data.ndim > 2:
        raise ValueError('Data with more than 2 dimensions not supported!')
    if ax is None:
        ax = plt.gca()

    n_pt = data.shape[0]
    time = np.arange(n_pt) / samplerate
    colors = sns.color_palette(color_palette, min(n_colors, n_pt))

    # Segment the line
    labels = np.zeros((n_pt,), dtype='int')
    size = (n_pt // n_colors) * n_colors # largest size we can evenly split into n_colors
    labels[:size] = np.repeat(range(n_colors), n_pt // n_colors)
    labels[size:] = n_colors - 1 # leftovers also get the last color
    times, _ = utils.segment_array(time, labels, duplicate_endpoints=True)
    lines, line_labels = utils.segment_array(data, labels, duplicate_endpoints=True)

    # Use linecollections to plot each channel of data
    labels = np.array(line_labels).astype(int)
    colors = [colors[i] for i in labels]
    for dim in range(data.shape[1]):
        segments = [np.vstack([t, l[:,dim]]).T for t, l in zip(times, lines)]
        lc = LineCollection(segments, colors=colors, **kwargs)
        ax.add_collection(lc)
        
    ax.margins(0.05) # add_collections doesn't autoscale
    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Voltage (V)')

def plot_freq_domain_amplitude(data, samplerate, ax=None, rms=False):
    '''
    Plots a amplitude spectrum of each channel on the given axis. Just need to input time series
    data and this will calculate and plot the amplitude spectrum. 

    Example:

        Plot 50 and 100 Hz sine wave amplitude spectrum. 

        .. code-block:: python

            data = np.sin(np.pi*np.arange(1000)/10) + np.sin(2*np.pi*np.arange(1000)/10)
            samplerate = 1000
            plot_freq_domain_amplitude(data, samplerate) # Expect 100 and 50 Hz peaks at 1 V each

        .. image:: _images/freqdomain.png

    Args:
        data (nt, nch): timeseries data in volts, can also be a single channel vector
        samplerate (float): sampling rate of the data
        ax (pyplot axis, optional): where to plot
        rms (bool, optional): compute root-mean square amplitude instead of peak amplitude
    '''
    if ax is None:
        ax = plt.gca()
    non_negative_freq, data_ampl = analysis.calc_freq_domain_amplitude(data, samplerate, rms)
    for ch in range(np.shape(data_ampl)[1]):
        ax.semilogx(non_negative_freq, data_ampl[:,ch])
    ax.set_xlabel('Frequency (Hz)')
    if rms:
        ax.set_ylabel('RMS amplitude (V)')
    else:
        ax.set_ylabel('Peak amplitude (V)')

def get_data_map(data, x_pos, y_pos):
    '''
    Organizes data according to the given x and y positions

    Args:
        data (nch): list of values
        x_pos (nch): list of x positions
        y_pos (nch): list of y positions

    Returns:
        (m,n array): map of the data on the grid defined by x_pos and y_pos
    '''
    data = np.reshape(data, -1)

    X = np.unique(x_pos)
    Y = np.unique(y_pos)
    nX = len(X)
    nY = len(Y)

    # Order Y into rows and X into columns
    data_map = np.empty((nY, nX), dtype=data.dtype)
    data_map[:] = np.nan
    for data_idx in range(len(data)):
        xid = np.where(X == x_pos[data_idx])[0]
        yid = np.where(Y == y_pos[data_idx])[0]
        data_map[yid, xid] = data[data_idx]

    return data_map


def calc_data_map(data, x_pos, y_pos, grid_size, interp_method='nearest', threshold_dist=None):
    '''
    Turns scatter data into grid data by interpolating up to a given threshold distance.

    Example:
        Make a plot of a 10 x 10 grid of increasing values with some missing data.
        
        .. code-block:: python

            data = np.linspace(-1, 1, 100)
            x_pos, y_pos = np.meshgrid(np.arange(0.5,10.5),np.arange(0.5, 10.5))
            missing = [0, 5, 25]
            data_missing = np.delete(data, missing)
            x_missing = np.reshape(np.delete(x_pos, missing),-1)
            y_missing = np.reshape(np.delete(y_pos, missing),-1)

            interp_map, xy = calc_data_map(data_missing, x_missing, y_missing, [10, 10], threshold_dist=1.5)
            plot_spatial_map(interp_map, xy[0], xy[1])

        .. image:: _images/posmap_calcmap.png

    Args:
        data (nch): list of values
        x_pos (nch): list of x positions
        y_pos (nch): list of y positions
        grid_size (tuple): number of points along each axis
        interp_method (str): method used for interpolation
        threshold_dist (float): distance to neighbors before disregarding a point on the image

    Returns:
        tuple: tuple containing:
            | *data_map (grid_size array, e.g. (16,16)):* map of the data on the given grid
            | *xy (grid_size array, e.g. (16,16)):* new grid positions to use with this map

    '''
    extent = [np.min(x_pos), np.max(x_pos), np.min(y_pos), np.max(y_pos)]

    x_spacing = (extent[1] - extent[0]) / (grid_size[0] - 1)
    y_spacing = (extent[3] - extent[2]) / (grid_size[1] - 1)
    xy = np.vstack((x_pos, y_pos)).T
    xq, yq = np.meshgrid(np.arange(extent[0], extent[0] + x_spacing * grid_size[0], x_spacing),
                         np.arange(extent[2], extent[2] + y_spacing * grid_size[1], y_spacing))
    
    # Remove nan values
    non_nan = np.logical_not(np.isnan(data))
    data = data[non_nan]
    xy = xy[non_nan]
    
    # Interpolate
    new_xy = (np.reshape(xq, -1), np.reshape(yq, -1))
    X = griddata(xy, data, new_xy, method=interp_method, rescale=False)

    # Construct kd-tree, functionality copied from scipy.interpolate
    tree = cKDTree(xy)
    xi = _ndim_coords_from_arrays((np.reshape(xq, -1), np.reshape(yq, -1)))
    dists, indexes = tree.query(xi)

    # Mask values with distances over the threshold with NaNs
    if threshold_dist:
        X[dists > threshold_dist] = np.nan

    data_map = np.reshape(X, grid_size)
    return data_map, new_xy


def plot_spatial_map(data_map, x, y, alpha_map=None, ax=None, cmap='bwr', nan_color='black', clim=None):
    '''
    Wrapper around plt.imshow for spatial data

    Args:
        data_map ((2,n) array): map of x,y data
        x (list): list of x positions
        y (list): list of y positions
        alpha_map ((2,n) array): map of alpha values (optional, default alpha=1 everywhere). If the alpha
            values are outside of the range (0,1) they will be scaled automatically.
        ax (int, optional): axis on which to plot, default gca
        cmap (str, optional): matplotlib colormap to use in image. default 'bwr'
        nan_color (str, optional): color to plot nan values, or None to leave them invisible. default 'black'
        clim ((2,) tuple): (min, max) to set the c axis limits. default None, show the whole range

    Returns:
        mappable: image object which you can use to add colorbar, etc.

    Examples:
        
        Make a plot of a 10 x 10 grid of increasing values with some missing data.
        
        .. code-block:: python
        
            data = np.linspace(-1, 1, 100)
            x_pos, y_pos = np.meshgrid(np.arange(0.5,10.5),np.arange(0.5, 10.5))
            missing = [0, 5, 25]
            data_missing = np.delete(data, missing)
            x_missing = np.reshape(np.delete(x_pos, missing),-1)
            y_missing = np.reshape(np.delete(y_pos, missing),-1)

            data_map = get_data_map(data_missing, x_missing, y_missing)
            plot_spatial_map(data_map, x_missing, y_missing)

        .. image:: _images/posmap.png

        Make the same image but include a transparency layer

        .. code-block:: python

            data = np.linspace(-1, 1, 100)
            x_pos, y_pos = np.meshgrid(np.arange(0.5,10.5),np.arange(0.5, 10.5))
            missing = [0, 5, 25]
            data_missing = np.delete(data, missing)
            x_missing = np.reshape(np.delete(x_pos, missing),-1)
            y_missing = np.reshape(np.delete(y_pos, missing),-1)
            data_map = get_data_map(data_missing, x_missing, y_missing)
            plot_spatial_map(data_map, x_missing, y_missing, alpha_map=data_map)

        .. image:: _images/posmap_alphamap.png

    '''
    # Calculate the proper extents
    if data_map.size > 1:
        extent = [np.min(x), np.max(x), np.min(y), np.max(y)]
        x_spacing = (extent[1] - extent[0]) / (data_map.shape[0] - 1)
        y_spacing = (extent[3] - extent[2]) / (data_map.shape[1] - 1)
        extent = np.add(extent, [-x_spacing / 2, x_spacing / 2, -y_spacing / 2, y_spacing / 2])
    else:
        extent = [np.min(x) - 0.5, np.max(x) + 0.5, np.min(y) - 0.5, np.max(y) + 0.5]

    # Set the 'bad' color to something different
    cmap = copy.copy(plt.get_cmap(cmap))
    if nan_color:
        cmap.set_bad(color=nan_color)
    
    # If an alpha map is present, make an rgba image
    if alpha_map is not None:
        if clim is None:
            clim = (np.nanmin(data_map), np.nanmax(data_map))
        norm = cm.colors.Normalize(*clim)
        scalarMap = cm.ScalarMappable(norm=norm, cmap=cmap)
        data_map = scalarMap.to_rgba(data_map)

        # Apply the alpha map after scaling from 0 to 1
        alpha_range = np.nanmax(alpha_map) - np.nanmin(alpha_map)
        if alpha_range > 1 or np.nanmax(alpha_map) > 1 or np.nanmin(alpha_map) < 0:
            alpha_map = (alpha_map - np.nanmin(alpha_map)) / alpha_range
        alpha_map[np.isnan(alpha_map)] = 0
        data_map[:,:,3] = alpha_map
        
    # Plot
    if ax is None:
        ax = plt.gca()
    image = ax.imshow(data_map, cmap=cmap, origin='lower', extent=extent)
    ax.set_xlabel('x position')
    ax.set_ylabel('y position')

    return image

def plot_ECoG244_data_map(data, bad_elec=[], interp=True, cmap='bwr', theta=0, ax=None, **kwargs):
    '''
    Plot a spatial map of data from an ECoG244 electrode array from the Viventi lab.

    Args:
        data ((256,) array): values from the ECoG array to plot in 2D
        bad_elec (list, optional): channels to remove from the plot. Defaults to [].
        interp (bool, optional): flag to include 2D interpolation of the result. Defaults to True.
        cmap (str, optional): matplotlib colormap to use in image. Defaults to 'bwr'.
        theta (float): rotation (in degrees) to apply to positions. rotations are applied clockwise, 
            e.g., theta = 90 rotates the map clockwise by 90 degrees, -90 rotates the map anti-clockwise 
            by 90 degrees. Default 0.
        ax (pyplot.Axes, optional): axis on which to plot. Defaults to None.
        kwargs (dict): dictionary of additional keyword argument pairs to send to calc_data_map and plot_spatial_map.

    Returns:
        pyplot.Image: image returned by pyplot.imshow. Use to add colorbar, etc.

    Examples:

        .. code-block:: python

            data = np.linspace(-1, 1, 256)
            missing = [0, 5, 25]
            plt.figure()
            plot_ECoG244_data_map(data, bad_elec=missing, interp=False, cmap='bwr', ax=None)
            # Here the missing electrodes (in addition to the ones
            # undefined by the channel mapping) should be visible in the map.

            plt.figure()
            plot_ECoG244_data_map(data, bad_elec=missing, interp=False, cmap='bwr', ax=None, nan_color=None)
            # Now we make the missing electrodes transparent

            plt.figure()
            plot_ECoG244_data_map(data, bad_elec=missing, interp=True, cmap='bwr', ax=None)
            # Missing electrodes should be filled in with linear interp.

    '''
    
    if ax is None:
        ax = plt.gca()
    
    # Load the signal path files
    elec_pos, acq_ch, elecs = aodata.load_chmap(drive_type='ECoG244', theta=theta)

    # Remove bad electrodes
    bad_ch = acq_ch[np.isin(elecs, bad_elec)]-1
    data[bad_ch] = np.nan
        
    # Interpolate or directly compute the map
    if interp:
        interp_kwargs = {k: v for k, v in kwargs.items() if k in ['interp_method', 'threshold_dist']}
        data_map, xy = calc_data_map(data[acq_ch-1], elec_pos[:,0], elec_pos[:,1], (16, 16), **interp_kwargs)
    else:
        data_map = get_data_map(data[acq_ch-1], elec_pos[:,0], elec_pos[:,1])
        xy = [elec_pos[:,0], elec_pos[:,1]]

    # Plot
    plot_kwargs = {k: v for k, v in kwargs.items() if k in ['alpha_map', 'nan_color', 'clim']}
    im = plot_spatial_map(data_map, xy[0], xy[1], cmap=cmap, ax=ax, **plot_kwargs)
    return im

def annotate_spatial_map(elec_pos, text, color, fontsize=6, ax=None, **kwargs):
    '''
    Simple wrapper around plt.annotate() to add text annotation to a 2d position. 

    Args:
        elec_pos ((x,y) tuple): position where text should be placed on 2d plot
        text (str): annotation text
        color (plt.Color): the color to make the text
        fontsize (int, optional): the fontsize to make the text. Defaults to 6.
        ax (pyplot.Axes, optional): axis on which to plot. Defaults to None.
        kwargs (dict): additional keyword arguments to pass to plt.annotate()

    Returns:
        plt.Annotation: annotation object
    '''
    if ax is None:
        ax = plt.gca()
    return ax.annotate(text, elec_pos, color=color, fontsize=fontsize, ha='center', va='center', **kwargs)
    
def annotate_spatial_map_channels(acq_idx=None, acq_ch=None, drive_type='ECoG244', theta=0, color='k', fontsize=6, 
                                  ax=None, **kwargs):
    '''
    Given acq_idx (indices) or acq_ch (channel numbers), prints either indices or channel numbers
    on top of a spatial map.

    Args:
        acq_idx ((nacq,) array or list, optional): If provided, specifies the acquisition indices to
            be annotated. If neither acq_idx nor acq_ch is provided, all channel numbers will be 
            annotated by default.
        acq_ch ((nacq,) array or list, optional): If provided, specifies the acquisition channel numbers to
            be annotated. If neither acq_idx nor acq_ch is provided, all channel numbers will be 
            annotated by default.
        drive_type (str, optional): Drive type of the channels to plot. See :func:`aopy.data.base.load_chmap`.
        color (str, optional): color to display the channels. Default 'k'.
        fontsize (int, optional): the fontsize to make the text. Defaults to 6.
        print_zero_index (bool, optional): if True (the default), prints channel numbers indexed by 0. 
            Otherwise prints directly from the channel map (which should use 1-indexing).
        ax (pyplot.Axes, optional): axis on which to plot. Defaults to None.
        kwargs (dict): additional keyword arguments to pass to plt.annotate()

    Example:

        .. code-block:: python

            aopy.visualization.plot_ECoG244_data_map(np.zeros(256,), cmap='Greys')
            aopy.visualization.annotate_spatial_map_channels(drive_type='ECoG244', color='k')
            aopy.visualization.annotate_spatial_map_channels(drive_type='Opto32', color='b')
            plt.axis('off')

        .. image:: _images/ecog244_opto32.png

    Note: 
        The acq_ch returned from `func::aopy.data.load_chmap` are generally 1-indexed lists of acquisition 
        channels connected to electrodes. In python, however, the acquisition indices start at 0, so we
        give the option to select channels based on either an index (acq_idx) or a channel number (acq_ch).
    '''
    if ax is None:
        ax = plt.gca()
    if acq_idx is not None and acq_ch is not None:
        raise ValueError("Please specify only one of acq_idx or acq_ch.")
    if acq_idx is not None:
        acq_ch = np.array(acq_idx)+1 # Change from index to ch numbers
        print("Annotating acquisition indices")
    else:
        print("Annotating acquisition channel numbers")

    # Get channel map (overwrite acq_ch if it was supplied to get the correct shape acq_ch)
    elec_pos, acq_ch, elecs = aodata.load_chmap(drive_type, acq_ch, theta)

    # Annotate each channel
    if isinstance(color, str) or len(color) < len(elec_pos):
        color = np.repeat(np.array(color), len(elec_pos))
    for pos, ch, color in zip(elec_pos, acq_ch, color):
        if acq_idx is not None:
            ch = ch - 1 # change back from channel numbers to indices
        annotate_spatial_map(pos, ch, color, fontsize, ax, **kwargs)

def plot_image_by_time(time, image_values, ylabel='trial', cmap='bwr', ax=None):
    '''
    Makes an nt x ntrial image colored by the timeseries values. 

    Example:
        ::

            time = np.array([-2, -1, 0, 1, 2, 3])
            data = np.array([[0, 0, 1, 1, 0, 0],
                            [0, 0, 0, 1, 1, 0]]).T
            plot_image_by_time(time, data)
            filename = 'image_by_time.png'

        .. image:: _images/image_by_time.png

    Args:
        time (nt): time vector to plot along the x axis
        image_values (nt, [nch or ntr]): time-by-trial or time-by-channel data
        ylabel (str, optional): description of the second axis of image_values. Defaults to 'trial'.
        cmap (str, optional): colormap with which to display the image. Defaults to 'bwr'.
        ax (pyplot.Axes, optional): Axes object on which to plot. Defaults to None.

    Returns:
        pyplot.AxesImage: the image object returned by pyplot
    '''
    
    image_values = np.array(image_values)
    extent = [np.min(time), np.max(time), 0, image_values.shape[1]]

    # Plot
    if ax is None:
        ax = plt.gca()
    im = ax.imshow(image_values.T, cmap=cmap, origin='lower', extent=extent, aspect='auto', \
        resample=False, interpolation='none', filternorm=False)
    ax.set_xlabel('time (s)')
    ax.set_ylabel(ylabel)
    
    return im


def plot_raster(data, cue_bin=None, ax=None):
    '''
    Create a raster plot for binary input data and show the relative timing of an event with a vertical red line

    .. image:: _images/raster_plot_example.png

    Args:
        data (ntime, ncolumns): 2D array of data. Typically a time series of spiking events across channels or trials (not spike count- must contain only 0 or 1).
        cue_bin (float): time bin at which an event occurs. Leave as 'None' to only plot data. For example: Use this to indicate 'Go Cue' or 'Leave center' timing.
        ax (plt.Axis): axis to plot raster plot
        
    Returns:
        None: raster plot plotted in appropriate axis
    '''
    if ax is None:
        ax = plt.gca()

    ax.eventplot(data.T, color='black')
    
    if cue_bin is not None:
        ax.axvline(x=cue_bin, linewidth=2.5, color='r')

def set_bounds(bounds, ax=None):
    '''
    Sets the x, y, and z limits according to the given bounds

    Args:
        bounds (tuple): 6-element tuple describing (-x, x, -y, y, -z, z) cursor bounds
        ax (plt.Axis, optional): axis to plot the targets on
    '''
    if ax is None:
        ax = plt.gca()

    try:
        ax.set(xlim=(1.1 * bounds[0], 1.1 * bounds[1]),
               ylim=(1.1 * bounds[2], 1.1 * bounds[3]),
               zlim=(1.1 * bounds[4], 1.1 * bounds[5]))
    except:
        ax.set(xlim=(1.1 * bounds[0], 1.1 * bounds[1]),
               ylim=(1.1 * bounds[2], 1.1 * bounds[3]))


def plot_targets(target_positions, target_radius, bounds=None, alpha=0.5, 
                 origin=(0, 0, 0), ax=None, unique_only=True):    
    '''
    Add targets to an axis. If any targets are at the origin, they will appear 
    in a different color (magenta). Works for 2D and 3D axes

    Example:
        Plot four peripheral and one central target.
        ::
        
            target_position = np.array([
                [0, 0, 0],
                [1, 1, 0],
                [-1, 1, 0],
                [1, -1, 0],
                [-1, -1, 0]
            ])
            target_radius = 0.1
            plot_targets(target_position, target_radius, (-2, 2, -2, 2))

        .. image:: _images/targets.png

    Args:
        target_positions (ntarg, 3): array of target (x, y, z) locations
        target_radius (float): radius of each target
        bounds (tuple, optional): 6-element tuple describing (-x, x, -y, y, -z, z) cursor bounds
        origin (tuple, optional): (x, y, z) position of the origin
        ax (plt.Axis, optional): axis to plot the targets on
        unique_only (bool, optional): If True, function will only plot targets with unique positions (default: True)
    '''

    if unique_only:
        target_positions = np.unique(target_positions,axis=0)

    if isinstance(alpha,float):
        alpha = alpha * np.ones(len(target_positions))
    else:
        assert len(alpha) == len(target_positions), "list of alpha values must be equal in length to the list of targets."

    if ax is None:
        ax = plt.gca()

    if unique_only:
        target_positions = np.unique(target_positions,axis=0)

    for i in range(len(target_positions)):

        # Pad the vector to make sure it is length 3
        pos = np.zeros((3,))
        pos[:len(target_positions[i])] = target_positions[i]

        # Color according to its position
        if (pos == origin).all():
            target_color = 'm'
        else:
            target_color = 'b'

        plot_circles([pos], target_radius, target_color, bounds, alpha[i], ax, unique_only=False)

def plot_circles(circle_positions, circle_radius, circle_color='b', bounds=None, alpha=0.5, ax=None, unique_only=True):    
    '''
    Add circles to an axis. Works for 2D and 3D axes

    Args:
        circle_positions (ntarg, 3): array of target (x, y, z) locations
        circle_radius (float): radius of each target
        circle_color (str): color to draw circle - default is blue
        bounds (tuple, optional): 6-element tuple describing (-x, x, -y, y, -z, z) cursor bounds
        origin (tuple, optional): (x, y, z) position of the origin
        ax (plt.Axis, optional): axis to plot the targets on
        unique_only (bool, optional): If True, function will only plot targets with unique positions (default: True)
    '''

    if unique_only:
        circle_positions = np.unique(circle_positions,axis=0)

    if isinstance(alpha,float):
        alpha = alpha * np.ones(len(circle_positions))
    else:
        assert len(alpha) == len(circle_positions), "list of alpha values must be equal in length to the list of cricles."

    if ax is None:
        ax = plt.gca()

    for i in range(0, len(circle_positions)):

        # Pad the vector to make sure it is length 3
        pos = np.zeros((3,))
        pos[:len(circle_positions[i])] = circle_positions[i]

        # Plot in 3D or 2D
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        try:
            ax.set_zlabel('z')
            u = np.linspace(0, 2 * np.pi, 100)
            v = np.linspace(0, np.pi, 100)
            x = pos[0] + circle_radius * np.outer(np.cos(u), np.sin(v))
            y = pos[1] + circle_radius * np.outer(np.sin(u), np.sin(v))
            z = pos[2] + circle_radius * np.outer(np.ones(np.size(u)), np.cos(v))
            ax.plot_surface(x, y, z, alpha=alpha[i], color=circle_color)
            ax.set_box_aspect((1, 1, 1))
        except:
            target = plt.Circle((pos[0], pos[1]),
                                radius=circle_radius, alpha=alpha[i], color=circle_color)
            ax.add_artist(target)
            ax.set_aspect('equal', adjustable='box')
    if bounds is not None: set_bounds(bounds, ax)


def plot_trajectories(trajectories, bounds=None, ax=None, **kwargs):
    '''
    Draws the given trajectories, one at a time in different colors. Works for 2D and 3D axes

    Example:
        Two random trajectories.
        ::

            trajectories =[
                np.array([
                    [0, 0, 0],
                    [1, 1, 0],
                    [2, 2, 0],
                    [3, 3, 0],
                    [4, 2, 0]
                ]),
                np.array([
                    [-1, 1, 0],
                    [-2, 2, 0],
                    [-3, 3, 0],
                    [-3, 4, 0]
                ])
            ]
            bounds = (-5., 5., -5., 5., 0., 0.)
            plot_trajectories(trajectories, bounds)

        .. image:: _images/trajectories.png

    Args:
        trajectories (list): list of (n, 2) or (n, 3) trajectories where n can vary across each trajectory
        bounds (tuple, optional): 6-element tuple describing (-x, x, -y, y, -z, z) cursor bounds
        ax (plt.Axis, optional): axis to plot the targets on
        kwargs (dict): keyword arguments to pass to the plt.plot function
   '''
    if ax is None:
        ax = plt.gca()

    # Plot in 3D, fall back to 2D
    try:
        ax.set_zlabel('z')
        for path in trajectories:
            ax.plot(*path.T, **kwargs)
        ax.set_box_aspect((1, 1, 1))
    except:
        for path in trajectories:
            ax.plot(path[:, 0], path[:, 1], **kwargs)
        ax.set_aspect('equal', adjustable='box')

    if bounds is not None: set_bounds(bounds, ax)
    ax.set_xlabel('x')
    ax.set_ylabel('y')

def color_trajectories(trajectories, labels, colors, ax=None, **kwargs):
    '''
    Draws the given trajectories but with the color of each trajectory corresponding to its given label.
    Works for 2D and 3D axes

    Example:

        .. code-block:: python

            trajectories =[
                    np.array([
                        [0, 0, 0],
                        [1, 1, 0],
                        [2, 2, 0],
                        [3, 3, 0],
                        [4, 2, 0]
                    ]),
                    np.array([
                        [-1, 1, 0],
                        [-2, 2, 0],
                        [-3, 3, 0],
                        [-3, 4, 0]
                    ])
                ]
            labels = [0, 0, 1, 0]
            colors = ['r', 'b']
            color_trajectories(trajectories, labels, colors)

            .. image:: _images/color_trajectories_simple.png

            labels_list = [[0, 0, 1, 1, 1], [0, 0, 1, 1], [1, 1, 0, 0]]
            fig = plt.figure()
            color_trajectories(trajectories, labels_list, colors)

            .. image:: _images/color_trajectories_segmented.png

    Args:
        trajectories (ntrials): list of (n, 2) or (n, 3) trajectories where n can 
            vary across each trajectory
        labels (ntrials): integer array of labels for each trajectory. Basically an 
            index for each trajectory
        colors (ncolors): list of colors. A list of arrays containing the label for 
            each corresponding trajectory, or a list of lists where each sublist 
            corresponds to the label for each timepoint in the corresponding trajectory.

        ax (plt.Axis, optional): axis to plot the targets on
        **kwargs (dict): other arguments for plot_trajectories(), e.g. bounds
    '''

    # If the labels are in list of lists format, segment the trajectories accordingly
    if isinstance(labels[0], list) or isinstance(labels[0], np.ndarray):
        all_trajectories = []
        all_labels = []
        for t, l in zip(trajectories, labels):
            assert len(t) == len(l), "Input labels must be the same length as input trajectories"
            segmented_trajectories, segmented_labels = utils.segment_array(t, l, duplicate_endpoints=True)
            all_trajectories += segmented_trajectories
            all_labels += segmented_labels
        trajectories = all_trajectories
        labels = all_labels
    
    # Convert the labels to integers for indexing into the color list
    labels = np.array(labels).astype(int)

    # Initialize a cycler with the appropriate colors
    style = plt.cycler(color=[colors[i] for i in labels])
    if ax is None:
        ax = plt.gca()
    ax.set_prop_cycle(style)

    # Use the regular trajectory plotting function
    plot_trajectories(trajectories, ax=ax, **kwargs)

def gradient_trajectories(trajectories, n_colors=100, color_palette='viridis', bounds=None, ax=None, **kwargs):
    '''
    Draw trajectories with a gradient of color from start to end of each trajectory. 
    Works in 2D and 3D.

    Note: this function applies the gradient evenly across the timepoints of the trajectory. 
        It might be useful to use the sampling rate of the data instead of n_colors, so that
        the time axis is consistent across sampling rates. 

    Args:
        trajectories (ntrials): list of 2D or 3D trajectories, in x, y[, z] coordinates
        n_colors (int, optional): number of colors in the gradient. Default 100.
        color_palette (str, optional): colormap to use for the gradient. Default 'viridis'.
        bounds (tuple, optional): 6-element tuple describing (-x, x, -y, y, -z, z) axes bounds
        ax (plt.Axis, optional): axis to plot the targets on
        kwargs (dict): keyword arguments to pass to the LineCollection function (similar to plt.plot)

    Example:

        Cursor trajectories in 2D
        .. code-block:: python

            subject = 'beignet'
            te_id = 5974
            date = '2022-07-01'
            preproc_dir = data_dir
            traj, _ = aopy.data.get_kinematic_segments(preproc_dir, subject, te_id, date, [32], [81, 82, 83, 239])
            gradient_trajectories(traj[:3])
        
        .. image:: _images/gradient_trajectories.png

        Hand trajectories in 3D
        .. code-block:: python

            traj, _ = aopy.data.get_kinematic_segments(preproc_dir, subject, te_id, date, [32], [81, 82, 83, 239], datatype='hand')
            plt.figure()
            ax = plt.axes(projection='3d')
            gradient_trajectories(traj[:3], bounds=[-10,0,60,70,20,40], ax=ax)

        .. image:: _images/gradient_trajectories_3d.png

    Note:
        Automatic bounds aren't set in 3D plots. The best alternative is to first plot in 2D, then use
        those bounds to manually set the first 2 axes bounds for the 3D plot.
    '''

    if ax is None:
        ax = plt.gca()

    color_list = sns.color_palette(color_palette, n_colors)
    for traj in trajectories:
        n_pt = len(traj)

        if n_pt < n_colors:
            warnings.warn("Not enough datapoints to divide into n_colors!")

        # Assign labels to the trajectory according to color
        labels = np.zeros((n_pt,), dtype='int')
        size = (n_pt // n_colors) * n_colors # largest size we can evenly split into n_colors
        labels[:size] = np.repeat(range(n_colors), n_pt // n_colors)
        labels[size:] = n_colors - 1 # leftovers also get the last color
            
        # Split the labeled trajectories into segments with unique colors
        segments, labels = utils.segment_array(traj, labels, duplicate_endpoints=True)
        labels = np.array(labels).astype(int)
        colors = [color_list[i] for i in labels]

        # Plot as line collections in 3D, fall back to 2D
        try:
            ax.set_zlabel('z')
            segments = [np.vstack([s[:,0], s[:,1], s[:,2]]).T for s in segments]
            lc = Line3DCollection(segments, colors=colors, **kwargs)
            ax.add_collection(lc)
            ax.set_box_aspect((1, 1, 1))
        except:
            segments = [np.vstack([s[:,0], s[:,1]]).T for s in segments]
            lc = LineCollection(segments, colors=colors, **kwargs)
            ax.add_collection(lc)
                
    try:
        ax.set_zlabel('z')
        ax.set_box_aspect((1, 1, 1))
    except:
        ax.set_aspect('equal', adjustable='box')

    if bounds is not None: 
        set_bounds(bounds, ax)
    else:
        ax.margins(0.05) # The ax.add_collection() call doesn't automatically set margins

    ax.set_xlabel('x')
    ax.set_ylabel('y')        

def plot_sessions_by_date(trials, dates, *columns, method='sum', labels=None, ax=None):
    '''
    Plot session data organized by date and aggregated such that if there are multiple rows on 
    a given date they are combined into a single value using the given method. If the method
    is 'mean' then the values will be averaged for each day, for example for size of cursor. The 
    average is weighted by the number of trials in that session. If the  method is 'sum' then the 
    values will be added together on each day, for example for number of trials.
    
    Example:

        Plotting success rate averaged across days.

        .. code-block:: python

            from datetime import date, timedelta
            date = [date.today() - timedelta(days=1), date.today() - timedelta(days=1), date.today()]
            success = [70, 65, 65]
            trials = [10, 20, 10]

            fig, ax = plt.subplots(1,1)
            plot_sessions_by_date(trials, dates, success, method='mean', labels=['success rate'], ax=ax)
            ax.set_ylabel('success (%)')

        .. image:: _images/sessions_by_date.png
        
    Args:
        trials (nsessions):
        dates (nsessions): 
        *columns (nsessions): dataframe columns or numpy arrays to plot
        method (str, optional): how to combine data within a single date. Can be 'sum' or 'mean'.
        labels (list, optional): string label for each column to go into the legend
        ax (pyplot.Axes, optional): axis on which to plot
    '''
    dates = np.array(dates)
    first_day = np.min(dates)
    last_day = np.max(dates)
    plot_days = pd.date_range(start=first_day, end=last_day).to_list()
    n_columns = len(columns)
    n_days = len(plot_days)
    aggregate = np.zeros((n_columns, n_days))

    for idx_day in range(n_days):
        day = plot_days[idx_day]
        for idx_column in range(n_columns):
            values = np.array(columns[idx_column])[dates == day.date()]
            
            try:
                if method == 'sum':
                    if len(values) > 0:
                        aggregate[idx_column, idx_day] = np.sum(values)
                    else:
                        aggregate[idx_column, idx_day] = np.nan
                elif method == 'mean':
                    day_trials = np.array(trials)[dates == day.date()]
                    aggregate[idx_column, idx_day] = np.average(values, weights=day_trials)
                else:
                    raise ValueError("Unknown method for combining data")
            except:
                aggregate[idx_column, idx_day] = np.nan

    if ax == None:
        ax = plt.gca()
    for idx_column in range(n_columns):
        if hasattr(columns[idx_column], 'name'):
            ax.plot(plot_days, aggregate[idx_column,:], '.-', label=columns[idx_column].name)
        else:
            ax.plot(plot_days, aggregate[idx_column,:], '.-')
    ax.xaxis.set_major_locator(mdates.WeekdayLocator(byweekday=(mdates.MO, mdates.TU, mdates.WE, 
                                                                mdates.TH, mdates.FR, mdates.SA, mdates.SU)))
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%m-%d'))
    plt.setp(ax.get_xticklabels(), rotation=80)
    
    if labels:
        ax.legend(labels)
    else:
        ax.legend()

def plot_sessions_by_trial(trials, *columns, dates=None, smoothing_window=None, labels=None, ax=None, **kwargs):
    '''
    Plot session data by absolute number of trials completed. Optionally split up the sessions by date and
    apply smoothing to each day's data.
    
    Example:

        Plotting success rate over three sessions.

        .. code-block:: python

            success = [70, 65, 60]
            trials = [10, 20, 10]

            fig, ax = plt.subplots(1,1)
            plot_sessions_by_trial(trials, success, labels=['success rate'], ax=ax)
            ax.set_ylabel('success (%)')

        .. image:: _images/sessions_by_trial.png
        
    Args:
        trials (nsessions): number of trials in each session
        *columns (nsessions): dataframe columns or numpy arrays to plot
        dates (nsessions, optional): dataframe columns or numpy arrays of the date of each session
        smoothing_window (int, optional): number of trials to smooth. Default no smoothing.
        labels (list, optional): string label for each column to go into the legend
        ax (pyplot.Axes, optional): axis on which to plot
    '''
    if ax == None:
        ax = plt.gca()  

    date_chg = []
    if dates is not None:
        trial_dates = np.repeat(np.array(dates), trials)
        date_chg = np.insert(np.where(np.diff(trial_dates) > timedelta(0))[0] + 1, 0, 0)

    for idx_column in range(len(columns)):
        
        # Accumulate individual trials with the values given for each session
        values = np.array(columns[idx_column])
        trial_values = np.repeat(values, trials)
        
        # Apply smoothing
        if smoothing_window is not None and dates is not None:
            split = np.split(trial_values, date_chg[1:])
            split = [analysis.calc_rolling_average(s, window_size=smoothing_window, mode='nan') for s in split]
            trial_values = np.concatenate(split)
        elif smoothing_window is not None:
            trial_values = analysis.calc_rolling_average(trial_values, window_size=smoothing_window)
        
        # Plot with additional kwargs
        if hasattr(columns[idx_column], 'name'):
            ax.plot(trial_values, label=columns[idx_column].name, **kwargs)
        else:
            ax.plot(trial_values, **kwargs)

    # Add date labels
    for i in date_chg:
        date = trial_dates[i]
        ax.axvline(i, ymin=0, ymax=1, color='gray', alpha=0.5, linestyle='dashed')
        ax.text(i, 1, str(date), color='gray', rotation=90, ha='left', va='top', 
                transform=ax.get_xaxis_transform())

    ax.set_xlabel('trials')
    if labels is not None:
        ax.legend(labels)
    else:
        ax.legend()

def plot_events_time(events, event_timestamps, labels, ax=None, colors=['tab:blue','tab:orange','tab:green']):
    '''
    This function plots multiple different events on the same plot. The first event (item in the list)
    will be displayed on the bottom of the plot.
    
    .. image:: _images/events_time.png
    
    Args:
        events (list (nevents) of 1D arrays (ntime)): List of Logical arrays that denote when an event(for example, a reward) occurred during an experimental session. Each item in the list corresponds to a different event to plot. 
        event_timestamps (list (nevents) of 1D arrays ntime): List of 1D arrays of timestamps corresponding to the events list. 
        labels (list (nevents) of str) : Event names for each list item.
        ax (axes handle): Axes to plot
        colors (list of str): Color to use for each list item
    '''

    if ax is None:
        ax = plt.gca()

    n_events = len(events)
    for i in range(n_events):
        this_events = events[i]
        this_timestamps = event_timestamps[i]
        n_timebins = np.shape(this_events)[0]

        if n_events <= len(colors):
            this_color = colors[i]
            ax.step(this_timestamps, 0.9*(this_events)+i+0.1, where='post', c=this_color)
        else:
            ax.step(this_timestamps, 0.9*(this_events)+i+0.1, where='post')
    ax.set_yticks(np.arange(n_events)+0.5)
    ax.set_yticklabels(labels)

    ax.set_xlabel('Time (s)') 

def plot_waveforms(waveforms, samplerate, plot_mean=True, ax=None):
    '''
    This function plots the input waveforms on the same figure and can overlay the mean if requested

    .. image:: _images/waveform_plot_example.png

    Args:
        waveforms (nt, nwfs): Array of waveforms to plot
        samplerate (float): Sampling rate of waveforms to calculate time axis. [Hz]
        plot_mean (bool): Indicate if the mean waveform should be plotted. Defaults to plot mean.
        ax (axes handle): Axes to plot
    '''

    if ax is None:
        ax = plt.gca()
    
    time_axis = (1e6)*np.arange(waveforms.shape[0])/samplerate

    if plot_mean:
        ax.plot(time_axis, waveforms, color='black', alpha=0.5)
        mean_waveform = np.nanmean(waveforms, axis=1)
        ax.plot(time_axis, mean_waveform, color='red')
    else:
        ax.plot(time_axis, waveforms)

    ax.set_xlabel(r'Time ($\mu$s)')

def plot_tuning_curves(fit_params, mean_fr, targets, n_subplot_cols=5, ax=None):
    '''
    This function plots the tuning curves output from analysis.run_tuningcurve_fit overlaying the actual firing rate data.
    The dashed line is the model fit and the solid line is the actual data. 

    .. image:: _images/tuning_curves_plot.png

    Args:
        fit_params (nunits, 3): Model fit coefficients. Output from analysis.run_tuningcurve_fit or analysis.curve_fitting_func
        mean_fr (nunits, ntargets): The average firing rate for each unit for each target.
        target_theta (ntargets): Orientation of each target in a center out task [degrees]. Corresponds to order of targets in 'mean_fr'
        n_subplot_cols (int): Number of columns to plot in subplot. This function will automatically calculate the number of rows. Defaults to 5
        ax (axes handle): Axes to plot

    '''
    nunits = mean_fr.shape[0]
    n_subplot_rows = ((nunits-1)//n_subplot_cols)+1
    axinput = True

    if ax is None:
        fig, ax = plt.subplots(n_subplot_rows, n_subplot_cols)
        axinput = False
        
    nplots = n_subplot_rows*n_subplot_cols
    for iunit in range(nplots):
        if nunits > n_subplot_cols and n_subplot_cols!=1:
          nrow = iunit//n_subplot_cols
          ncol = iunit - (nrow*n_subplot_cols)
          # Remove axis that aren't used
          if iunit >= nunits:
            ax[nrow, ncol].remove()
          else:
            ax[nrow, ncol].plot(targets, mean_fr[iunit,:], 'b-', label='data')
            ax[nrow, ncol].plot(targets, analysis.curve_fitting_func(targets, fit_params[iunit, 0], fit_params[iunit, 1], fit_params[iunit,2]), 'b--', label='fit')
            ax[nrow, ncol].set_title('Unit ' +str(iunit))

        else:
          # Remove axis that aren't used
          if iunit >= nunits:
            ax[iunit].remove()
          else:
            ax[iunit].plot(targets, mean_fr[iunit,:], 'b-', label='data')
            ax[iunit].plot(targets, analysis.curve_fitting_func(targets, fit_params[iunit, 0], fit_params[iunit, 1], fit_params[iunit,2]), 'b--', label='fit')
            ax[iunit].set_title('Unit ' +str(iunit))

    if not axinput:
        fig.tight_layout()
        
def plot_boxplots(data, plt_xaxis, trendline=True, facecolor='gray', linecolor='k', box_width=0.5, ax=None):
    '''
    This function creates a boxplot for each column of input data. If the input data has NaNs, they are ignored.

    Args:
        data (ncol list or (m, ncol) array): Data to plot. A different boxplot is created for each entry of the list.
        plt_xaxis (ncol): X-axis locations or labels to plot the boxplot of each column
        trendline (bool): If a line should be used to connect boxplots
        facecolor (color): Color of the box faces. Can be any input that pyplot interprets as a color.
        linecolor (color): Color of the connecting lines.
        ax (axes handle): Axes to plot

    Examples:

        Using a rectangular array and numeric x-axis points.

        .. code-block:: python
            
            data = np.random.normal(0, 2, size=(20, 5))
            xaxis_pts = np.array([2,3,4,4.75,5.5])
            fig, ax = plt.subplots(1,1)
            plot_boxplots(data, xaxis_pts, ax=ax)

        .. image:: _images/boxplot_example.png

        Using a list of nonrectangular arrays with categorical x-axis points.

        .. code-block:: python
        
            data = [np.random.normal(0, 2, size=(10)), np.random.normal(0, 1, size=(20))]
            xaxis_pts = ['foo', 'bar']
            fig, ax = plt.subplots(1,1)
            plot_boxplots(data, xaxis_pts, ax=ax)

        .. image:: _images/boxplot_example_nonrectangular.png

    '''
    if ax is None:
        ax = plt.gca()

    # If data is 2D, turn the columns into lists
    if hasattr(data, 'ndim') and data.ndim == 2:
        data = [data[:,i] for i in range(data.shape[1])]

    # If data is a single column, make it a list
    try:
        int(data[0])
        data = [data]
    except:
        pass
        
    if trendline:
        ax.plot(plt_xaxis, [np.nanmedian(data[i]) for i in range(len(data))], color=facecolor)
    
    for featidx, ifeat in enumerate(plt_xaxis):
        temp_data = data[featidx]
        try:
            int(ifeat)
        except:
            ifeat = featidx
        ax.boxplot(temp_data[~np.isnan(temp_data)], 
            positions=np.array([ifeat]), patch_artist=True, widths=box_width, 
            boxprops=dict(facecolor=facecolor, color=linecolor), capprops=dict(color=linecolor),
            whiskerprops=dict(color=linecolor), flierprops=dict(color=facecolor, markeredgecolor=facecolor),
            medianprops=dict(color=linecolor))
    ax.set_xticklabels(plt_xaxis)

def advance_plot_color(ax, n):
    '''
    Utility to skip colors for the given axis.
    
    Args:
        ax (pyplot.Axes): specify which axis to advance the color
        n (int): how many colors to skip in the cycle

    Examples:

        Using advance_plot_color to skip the first color in the cycle.

        .. code-block:: python

            plt.subplots()
            aopy.visualization.advance_plot_color(plt.gca(), 1)
            plt.plot(np.arange(10), np.arange(10))

        .. image:: _images/advance_plot_color.png
        
    '''
    for _ in range(n):
        next(ax._get_lines.prop_cycler)

def reset_plot_color(ax):
    '''
    Utility to reset the color cycle on a given axis to the default.

    Args:
        ax (pyplot.Axes): specify which axis to reset the color

    Examples:

        Using reset_plot_color to reset the color cycle between calls to `plt.plot()`.
        
        .. code-block:: python

            plt.subplots()
            plt.plot(np.arange(10), np.ones(10))
            aopy.visualization.reset_plot_color(plt.gca())
            plt.plot(np.arange(10), 1 + np.ones(10))

        .. image:: _images/reset_plot_color.png
    '''
    ax.set_prop_cycle(None)

def profile_data_channels(data, samplerate, figuredir, **kwargs):
    """
    Runs `plot_channel_summary` and `combine_channel_figures` on all channels in a data array

    Args:
        data (nt, nch): numpy array of neural data
        samplerate (int): sampling rate of data
        figuredir (str): string indicating file path to desired save directory
        kwargs (**dict): keyword arguments to pass to plot_channel_summary()

    .. image:: _images/channel_profile_example.png
    
    """
    
    if not os.path.exists(figuredir):
        os.makedirs(figuredir)
    _, nch = data.shape
    
    for chidx in tqdm(range(nch)):
        chname = f'ch. {chidx+1}'
        fig = plot_channel_summary(data[:,chidx], samplerate, title=chname, **kwargs)
        fig.savefig(os.path.join(figuredir,f'ch_{chidx}.png'))
        
    combine_channel_figures(figuredir, nch=nch, figsize=kwargs.pop('figsize', (6,5)), dpi=kwargs.pop('dpi', 150))

    
def combine_channel_figures(figuredir, nch=256, figsize=(6,5), dpi=150):
    """
    Combines all channel figures in directory generated from plot_channel_summary

    Args:
        figuredir (str): path to directory of channel profile images
        nch (int, optional): number of channels from data array. Determines combined image layout. Defaults to 256.
        figsize (tuple, optional): (width, height) to pass to pyplot. Default (6, 5)
        dpi (int, optional): resolution to pass to pyplot. Default 150
    """
    
    assert os.path.exists(figuredir), f"Directory not found: {figuredir}"
    
    ncol = int(np.ceil(np.sqrt(nch))) # make things as square as possible
    nrow = int(np.ceil(nch/ncol))
    imgw = figsize[0] * dpi # I should get these from the individual files...
    imgh = figsize[1] * dpi
    
    grid = Image.new(mode='RGB', size=(ncol*imgw, nrow*imgh))
    
    print(f'profiling all {nch} channels...')
    for chidx in tqdm(range(nch)):
        figurefile = os.path.join(figuredir,f'ch_{chidx}.png')
        rowidx = chidx // ncol
        colidx = chidx % ncol
        if not os.path.exists(figurefile):
            continue
        else:
            with Image.open(figurefile) as img:
                grid.paste(img,box=(colidx*imgw, rowidx*imgh))
    
    grid.save(os.path.join(figuredir,'all_ch.png'),'png')


def plot_channel_summary(chdata, samplerate, nperseg=None, noverlap=None, trange=None, title=None, figsize=(6, 5), dpi=150, frange=(0, 80), cmap_lim=(0, 40)):
    """
    Plot time domain trace, spectrogram and normalized (z-scored) spectrogram. Computes spectrogram.
    
    ::

        ---------------
        | time series |
        |-------------|
        | spectrogram |
        |-------------|
        | norm sgram  |
        ---------------
    
    Args:
        chdata (nt,1): neural recording data from a given channel (lfp, ecog, broadband)
        samplerate (int): data sampling rate
        nperseg (int): length of each spectrogram window (in samples)
        noverlap (int): number of samples shared between neighboring spectrogram windows (in samples)
        trange (tuple, optional): (min, max) time range to display. Default show the entire time series
        title (str, optional): print a title above the timeseries data. Default None
        figsize (tuple, optional): (width, height) to pass to pyplot. Default (6, 5)
        dpi (int, optional): resolution to pass to pyplot. Default 150
        frange (tuple, optional): range of frequencies to display in spectrogram. Default (0, 80)
        cmap_lim (tuple, optional): clim to display in the spectrogram. Default (0, 40)

    Outputs:
        fig (Figure): Figure object
    """
    
    assert len(chdata.shape) < 2, "Input data array must be 1d"
    
    time = np.arange(len(chdata))/samplerate
    if trange is None:
        trange = (time[0], time[-1])
                                   
    if nperseg is None:
        nperseg = int(2*samplerate)
                                
    if noverlap is None:
        noverlap = int(1.5*samplerate)
    
    f_sg, t_sg, sgram = signal.spectrogram(
        chdata,
        fs=samplerate,
        nperseg=nperseg,
        noverlap=noverlap,
        detrend='linear'
    )
    log_sgram = np.log10(sgram)
    
    fig, ax = plt.subplots(3,1,figsize=figsize,dpi=dpi,constrained_layout=True,sharex=True)
    ax[0].plot(time, chdata)
    sg_pcm = ax[1].pcolormesh(t_sg,f_sg,10*log_sgram,vmin=cmap_lim[0],vmax=cmap_lim[1],shading='auto')
    ax[1].set_ylim(*frange)
    sg_cb = plt.colorbar(sg_pcm,ax=ax[1])
    sg_cb.ax.set_ylabel('dB$\mu$')
    sgn_pcm = ax[2].pcolormesh(t_sg,f_sg,zscore(log_sgram,axis=-1),vmin=-3,vmax=3,shading='auto',cmap='bwr')
    ax[2].set_ylim(*frange)
    sgn_cb = plt.colorbar(sgn_pcm,ax=ax[2])
    sgn_cb.ax.set_ylabel('z-scored dB$\mu$')
    ax[0].set_xlim(*trange)
    ax[0].set_ylabel('amp. ($\mu V$)')
    ax[1].set_ylabel('freq. (Hz)')
    ax[2].set_ylabel('freq. (Hz)')
    ax[2].set_xlabel('time (s)')
    ax[0].set_title(title)
    
    return fig


def plot_corr_over_elec_distance(elec_data, elec_pos, ax=None, **kwargs):
    '''
    Makes a plot of correlation vs electrode distance for the given data.
    
    Args:
        elec_data (nt, nelec): electrode data with nch corresponding to elec_pos
        elec_pos (nelec, 2): x, y position of each electrode
        ax (pyplot.Axes, optional): axis on which to plot
        kwargs (dict, optional): other arguments to supply to :func:`aopy.analysis.calc_corr_over_elec_distance`

    Example:
        Using the multichannel test data generator in utils, we get a phase-shifted sine wave in each channel. 
        Assigning each channel i to an electrode with position (i, 0), the correlation across distances looks like this:

        .. code-block:: python

            duration = 0.5
            samplerate = 1000
            n_channels = 30
            frequency = 100
            amplitude = 0.5
            acq_data = aopy.utils.generate_multichannel_test_signal(duration, samplerate, n_channels, frequency, amplitude)
            acq_ch = (np.arange(n_channels)+1).astype(int)
            elec_pos = np.stack((range(n_channels), np.zeros((n_channels,))), axis=-1)
            
            plt.figure()
            plot_corr_over_elec_distance(acq_data, acq_ch, elec_pos)

        .. image:: _images/corr_over_dist.png

    Updated:
        2024-03-13 (LRS): Changed input from acq_data and acq_ch to elec_data.
    '''
    if ax is None:
        ax = plt.gca()
    label = kwargs.pop('label', None)
    dist, corr = analysis.calc_corr_over_elec_distance(elec_data, elec_pos, **kwargs)
    ax.plot(dist, corr, label=label)
    ax.set_xlabel('binned electrode distance (cm)')
    ax.set_ylabel('correlation')
    ax.set_ylim(0,1)

def plot_corr_across_entries(preproc_dir, subjects, ids, dates, band=(70,200), taper_len=0.1, num_seconds=60, 
                             cmap='viridis', ax=None, remove_bad_ch=True, **bad_ch_kwargs):
    '''
    Plot the correlation vs electrode distance for each entry in the given list of subjects, ids, and dates.
    
    Args:
        preproc_dir (str): path to the preprocessed data directory
        subjects (list): list of subject names
        ids (list): list of te_ids
        dates (list): list of dates
        band (tuple, optional): frequency band to filter the data. Default (70, 200)
        taper_len (float, optional): length of taper to use in the filter. Default 0.1
        num_seconds (int, optional): number of seconds to use in the correlation calculation. Default 60
        cmap (str, optional): colormap to use for plotting. Default 'viridis'
        ax (pyplot.Axes, optional): axis on which to plot. Default current axis
        remove_bad_ch (bool, optional): whether to remove bad channels from the data. Default True
        bad_ch_kwargs (dict, optional): keyword arguments to pass to :func:`a
    
    Example:
        Plotting the correlation vs electrode distance for a few entries in the preprocessed data directory.

        .. image:: _images/corr_over_entries.png
    '''
    assert len(subjects) == len(ids) == len(dates), "Subjects, ids, and dates must be equal length"
    
    if ax is None:
        ax = plt.gca()    
    ax.set_prop_cycle('color', sns.color_palette(cmap, len(subjects)))
        
    for subject, te_id, date in zip(subjects, ids, dates):

        try:
            lfp_data, lfp_metadata = aodata.load_preproc_lfp_data(preproc_dir, subject, te_id, date)
            exp_data, exp_metadata = aodata.load_preproc_exp_data(preproc_dir, subject, te_id, date)
        except:
            print(f"Could not find data for entry {te_id} ({subject} on {date})")
            continue
        try:
            elec_pos, acq_ch, _ = aodata.load_chmap(exp_metadata['drmap_drive_type'])
        except:
            elec_pos, acq_ch, _ = aodata.load_chmap('ECoG244')

        samplerate = lfp_metadata['samplerate']
        short_data = lfp_data[:num_seconds*samplerate,acq_ch-1]
        filt_data = precondition.mt_bandpass_filter(short_data, band, taper_len, 
                                                            samplerate, verbose=False)
        
        if remove_bad_ch:
            bad_ch = preproc.quality.detect_bad_ch_outliers(filt_data, **bad_ch_kwargs)
            filt_data = filt_data[:,~bad_ch]
            elec_pos = elec_pos[~bad_ch]
        
        plot_corr_over_elec_distance(filt_data, elec_pos, label=date, ax=ax)

    leg = ax.legend(bbox_to_anchor = (1,1))
    for obj in leg.legend_handles:
        obj.set_linewidth(4.0)

def plot_tfr(values, times, freqs, cmap='plasma', logscale=False, ax=None, **kwargs):
    '''
    Plot a time-frequency representation of a signal.

    Args:
        values ((nt, nfreq) array): 
        times ((nt,) array): 
        freqs ((nfreq,) array): 
        cmap (str, optional): colormap to use for plotting
        logscale (bool, optional): apply a log scale to the color axis. Default False.
        ax (pyplot.Axes, optional): axes on which to plot. Default current axis.
        kwargs (dict, optional): other keyword arguments to pass to pyplot
        
    Returns:
        pyplot.Image: image object returned from pyplot.pcolormesh. Useful for adding colorbars, etc.
        
    Examples:
        
        .. code-block:: python

            fig, ax = plt.subplots(3,1,figsize=(4,6))

            samplerate = 1000
            data_200_hz = aopy.utils.generate_multichannel_test_signal(2, samplerate, 8, 200, 2)
            nt = data_200_hz.shape[0]
            data_200_hz[:int(nt/3),:] /= 3
            data_200_hz[int(2*nt/3):,:] *= 2

            data_50_hz = aopy.utils.generate_multichannel_test_signal(2, samplerate, 8, 50, 2)
            data_50_hz[:int(nt/2),:] /= 2

            data = data_50_hz + data_200_hz
            print(data.shape)
            aopy.visualization.plot_timeseries(data, samplerate, ax=ax[0])
            aopy.visualization.plot_freq_domain_amplitude(data, samplerate, ax=ax[1])

            freqs = np.linspace(1,250,100)
            coef = aopy.analysis.calc_cwt_tfr(data, freqs, samplerate, fb=10, f0_norm=1, verbose=True)
            t = np.arange(nt)/samplerate
            
            print(data.shape)
            print(coef.shape)
            print(t.shape)
            print(freqs.shape)
            pcm = aopy.visualization.plot_tfr(abs(coef[:,:,0]), t, freqs, 'plasma', ax=ax[2])

            fig.colorbar(pcm, label='Power', orientation = 'horizontal', ax=ax[2])
            
        .. image:: _images/tfr_cwt_50_200.png


    See Also:
        :func:`~aopy.analysis.calc_cwt_tfr`
    '''
    
    if ax == None:
        ax = plt.gca()
        
    if logscale:
        pcm = ax.pcolormesh(times, freqs, np.log10(values), cmap=cmap, **kwargs)
    else:
        pcm = ax.pcolormesh(times, freqs, values, cmap=cmap, **kwargs)
    pcm.set_edgecolor('face')
    
    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Frequency (Hz)')

    return pcm

def get_color_gradient_RGB(npts, end_color, start_color=[1,1,1]):
    '''
    This function outputs an ordered list of RGB colors that are linearly spaced between white and the input color. See also sns.color_palette for a gradient of RGB values within a Seaborn color palette.

    Examples:
        
        .. code-block:: python

                npts = 200
                x = np.linspace(0, 2*np.pi, npts)
                y = np.sin(x)
                fig, ax = plt.subplots()
                ax.scatter(x, y, c=get_color_gradient(npts, 'g', [1,0,0]))
    
        .. image:: _images/color_gradient_example.png

    Args:
        npts (int): How many different colors are part of the gradient
        end_color (str or list): Color that ends the gradient. Can be any matplotlib color or specific RGB values.
        start_color (str or list): Color that ends the gradient. Can be any matplotlib color or specific RGB values. Defaults to white.

    Returns:
        (npts, 3): An array with linearly spaced colors from the start to end
    '''
    rgb_end = matplotlib.colors.to_rgb(end_color)
    rgb_start = matplotlib.colors.to_rgb(start_color)
    ct = np.zeros((npts, 3))
    ct[:,0] = np.flip(np.linspace(rgb_end[0], rgb_start[0], npts))
    ct[:,1] = np.flip(np.linspace(rgb_end[1], rgb_start[1], npts))
    ct[:,2] = np.flip(np.linspace(rgb_end[2], rgb_start[2], npts))
    return ct

def plot_laser_sensor_alignment(sensor_volts, samplerate, stim_times, ax=None):
    '''
    Plot laser sensor data aligned to the stimulus times. Useful to debug laser timing issues to
    make sure the laser is actually on when you think it is.

    Args:
        sensor_volts ((nstim,) float array): laser sensor data
        samplerate (float): sampling rate of the sensor data
        stim_times ((nstim,) array): times at which the laser was turned on
        ax (pyplot.Axes, optional): axes on which to plot. Default current axis.
        kwargs (dict, optional): other keyword arguments to pass to pyplot
    
    Returns:
        pyplot.Image: image object returned from pyplot.pcolormesh. Useful for adding colorbars, etc.

    Examples:
        .. image:: _images/laser_sensor_alignment.png
    '''
    if ax is None:
        ax = plt.gca()
    time_before = 0.1 # seconds
    time_after = 0.1 # seconds
    analog_erp = analysis.calc_erp(sensor_volts, stim_times, time_before, time_after, samplerate)
    t = 1000*(np.arange(analog_erp.shape[0])/samplerate - time_before) # milliseconds
    im = plot_image_by_time(t, analog_erp[:,0,:], ylabel='trials')
    plt.xlabel('time (ms)')
    plt.title('laser sensor aligned')
    return im



# postproc.py
# Code for post-processing neural data, including separating neural features such as 
# LFP bands or spikes detection / binning

import numpy as np
import math
import warnings
from .. import precondition
from ..preproc.base import interp_timestamps2timeseries, get_data_segments, get_trial_segments, trial_align_data
from ..utils import derivative
from ..data import load_preproc_exp_data, load_preproc_eye_data, load_preproc_lfp_data
from .. import data

def translate_spatial_data(spatial_data, new_origin):
    '''
    Shifts 2D or 3D spatial data to a new location.

    Args:
        spatial_data (nt, ndim): Spatial data in 2D or 3D
        new_origin (ndim): Location of point that will become the origin in cartesian coordinates

    Returns:
        new_spatial_data (nt, ndim): new reach trajectory translated to the new origin
    '''
    new_spatial_data = np.subtract(spatial_data, new_origin)

    return new_spatial_data

def rotate_spatial_data(spatial_data, new_axis, current_axis):
    '''
    Rotates data about the origin into a new coordinate system based on the relationship between
    'new_axis' and 'current_axis'. If 'current_axis' and 'new_axis' point in 
    the same direction, the code will return 'spatial_data' with a warning that the vectors point in
    the same direction.
    
    This function was written to rotate spatial data but can be applied to other data of similar form.

    Args:
        spatial_data (nt, ndim): Array of spatial data in 2D or 3D
        new_axis (ndim): vector pointing along the desired orientation of the data
        current_axis (ndim): vector pointing along the current orientation of the dat

    Returns:
        output_spatial_data (nt, ndim): new reach trajectory rotated to the new axis

    Updates: July 2023 : updated function to work when new_axis is not a unit vector
    '''

    # Check if input data is a single point and enfore that it is a row vector
    if len(spatial_data.shape) == 1:
        spatial_data.shape = (1,len(spatial_data))    

    # Initialize output array
    output_spatial_data = np.empty((spatial_data.shape[0], 3))

    # Check for a 2D or 3D trajectory and convert to 3D points
    if spatial_data.shape[1] == 2:
        spatial_data3d = np.concatenate((spatial_data, np.zeros((spatial_data.shape[0],1))), axis = 1)
        new_axis3d = np.concatenate((new_axis, np.array([0])))
        current_axis3d = np.concatenate((current_axis, np.array([0])))
    elif spatial_data.shape[1] == 3:
        spatial_data3d = spatial_data
        new_axis3d = new_axis
        current_axis3d = current_axis

    # Calcualte angle between 'new_axis3d' and target trajectory via dot product
    angle = np.arccos(np.dot(new_axis3d, current_axis3d)/(np.linalg.norm(new_axis3d)*np.linalg.norm(current_axis3d)))

    # If angle is 0, return the original data and warn
    if np.isclose(angle, 0, atol = 1e-8):
        warnings.warn("Starting and desired vector are the same. No rotation applied")
        output_spatial_data = spatial_data3d
        return output_spatial_data

    # If the angle is exactly 180 degrees, slightly nudge the starting vector by 1e-7
    elif (spatial_data.shape[1] == 2) and np.isclose(angle, np.pi, atol = 1e-8):
        current_axis3d = current_axis3d.astype('float64')
        current_axis3d[:2] += 1e-7
    elif np.isclose(angle, np.pi, atol = 1e-8):
        raise ValueError("180 degree rotation cannot be resolved in 3D")

    # Calculate unit vector axis to rotate about via cross product
    rotation_axis = np.cross(current_axis3d, new_axis3d)
    unit_rotation_axis = rotation_axis/np.linalg.norm(rotation_axis)

    # Calculate quaternion
    # https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Using_quaternions_as_rotations
    qr = np.cos(angle/2)
    qi = np.sin(angle/2)*unit_rotation_axis[0]
    qj = np.sin(angle/2)*unit_rotation_axis[1]
    qk = np.sin(angle/2)*unit_rotation_axis[2]

    # Convert quaternion to rotation matrix
    # https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Quaternion-derived_rotation_matrix
    rotation_matrix = np.array([[1-2*(qj**2 + qk**2), 2*(qi*qj - qk*qr), 2*(qi*qk + qj*qr)],
                                [2*(qi*qj + qk*qr), 1-2*(qi**2 + qk**2), 2*(qj*qk - qi*qr)],
                                [2*(qi*qk - qj*qr), 2*(qj*qk + qi*qr), 1-2*(qi**2 + qj**2)]])

    # Apply rotation matrix to each point in the trajectory
    output_spatial_data = rotation_matrix @ spatial_data3d.T
    output_spatial_data = output_spatial_data.T

    # Check that we did the correct rotation
    output_axis = rotation_matrix @ current_axis3d
    output_axis = output_axis/np.linalg.norm(output_axis)
    assert np.allclose(output_axis, new_axis3d/np.linalg.norm(new_axis3d), atol=1e-4), "Something went wrong!"

    # Return trajectories the the same dimensions as the input
    if spatial_data.shape[1] == 2:
        return output_spatial_data[:,:2]
    elif spatial_data.shape[1] == 3:
        return output_spatial_data

def calc_reward_intervals(timestamps, values):
    '''
    Given timestamps and values corresponding to reward times and reward state, calculate the
    intervals (start, end) during which the reward was active

    Args:
        timestamps (nt): when the reward transitioned state
        values (nt): what the state was at each corresponding timestamp

    Returns:
        (nt/2): during which the reward was active
    '''
    reward_ts_on = timestamps[values == 1]
    reward_ts_off = timestamps[values == 0]
    if len(reward_ts_on) == len(reward_ts_off):
        return list(zip(reward_ts_on, reward_ts_off))
    else:
        raise ValueError("Invalid reward timestamps or values")

def get_trial_targets(trials, targets):
    '''
    Organizes targets from each trial into a trial array of targets. Essentially reshapes the array,
    but sometimes? there can be more or fewer targets in certain trials than in others

    Args:
        trials (ntargets): trial number for each target presented
        targets (ntargets, 3): target locations
    
    Returns:
        (ntrials list of (ntargets, 3)): list of targets in each trial
    ''' 
    n_trials = np.max(trials) + 1
    trial_targets = [[] for _ in range(n_trials)]
    for idx in range(len(trials)):
        trial = trials[idx]
        trial_targets[trial].append(targets[idx])
    return trial_targets

def get_minimum_trials_per_target(target_idx, cond_mask=None):
    '''
    Get the minimum number of trials per target after restricting trials
    
    Args:
        target_index (ntr): target index
        cond_mask (ntr): boolean array to remove trials
        
    Returns:
        (int): minimum number of trials per target
        
    '''
    
    if cond_mask is None:
        min_trial = min([sum(target_idx == itarget) for itarget in np.unique(target_idx)])
    else:
        min_trial = min([sum(target_idx[cond_mask] == itarget) for itarget in np.unique(target_idx)])
    
    return min_trial

def get_conditioned_trials_per_target(target_idx, trials_per_cond, cond_mask=None, replacement=False, seed=None):
    '''
    Get trial index to choose the same number of trials per target in removing trials by a certain condition.
    The trial index is evenly aligned like [1,2,3,1,2,3,1,2,3,...].
    trials_per_cond can be taken from 'get_minimum_trials_per_target' function.
    
    Args:
        target_index (ntr): target index
        trials_per_cond (int): minimum trial across conditions to get the same number of trials per condition
        cond_mask (ntr): boolean array to remove trials
        replacement (bool): whether to allow replacement in choosing trials. This can be used for bootstrapping.
        seed (int): random seed
        
    Returns:
        (ntr): trial index to extract the same number of conditioned trials for each target
        
    '''
    if seed is not None:
        np.random.seed(seed)
        
    # Get trial index to get the same number of trials per target
    for idx, itarget in enumerate(np.unique(target_idx)):
        if cond_mask is None:
            trial_mask_targ = np.where(target_idx == itarget)[0]
        else:
            trial_mask_targ = np.where(cond_mask * (target_idx == itarget))[0] 
                   
        if trial_mask_targ.size:
            if idx == 0:
                trial_mask = np.random.choice(trial_mask_targ, trials_per_cond, replace=replacement)
            else:
                trial_mask = np.vstack([trial_mask, np.random.choice(trial_mask_targ, trials_per_cond, replace=replacement)]) # conditions x trials shape

    # reshape using 'F' so that trial index would be aligned in the pseudorandom order
    trial_mask = trial_mask.reshape(-1, order='F')
    
    return trial_mask

def get_relative_point_location(ref_point_pos, new_point_pos):
    '''
    This function calculates the relative location (angle and position) of a point compared to a reference point.
    Assumes angle 0 starts from the direciton of vector [1, 0].
    This function is specific to the 2D case at the moment but can handle a single point entry or multiple.

    Args:
        ref_point_pos (nxpts, nypts): Point position of reference point in spatial coordinates
        new_point_pos (nxpts, nypts): Point position of new point in spatial coordinates

    Returns:
        A tuple containing:
            | **relative_new_point_angle (float):** Absolute angle between cursor and target position [rad]
            | **relative_new_point_pos (float):** Absolute position of the target relative to the cursor
    '''
    # Handle single point case
    relative_new_point_pos = new_point_pos - ref_point_pos

    if len(ref_point_pos.shape) == 1 or relative_new_point_pos.shape[1] == 1:
        relative_new_point_angle = np.arctan2(relative_new_point_pos[1], relative_new_point_pos[0])
        if relative_new_point_angle < 0:
            relative_new_point_angle = 2*np.pi + relative_new_point_angle
    
    # Handle multi-point case
    else:
        relative_new_point_angle = np.arctan2(relative_new_point_pos[:,1], relative_new_point_pos[:,0])
        relative_new_point_angle_mask = relative_new_point_angle < 0
        relative_new_point_angle[relative_new_point_angle_mask] = 2*np.pi + relative_new_point_angle[relative_new_point_angle_mask]

    return relative_new_point_angle, relative_new_point_pos

def get_inst_target_dir(trial_aligned_pos, targetpospertrial):
    '''
    This function calculates the instantaneous direction from the cursor to the target at each time point across each trial.
    This function is specific to the 2D case at the moment with X coordinates being the horizontal dimension and Y coordinates being the vertical dimension. 

    Args:
         trial_aligned_pos (ntime, ntrials, 2): Position of the cursor in X and Y coordinates.trial_aligned_pos[:,:,0] corresponds to the X cursor positions and trial_aligned_pos[:,:,1] to the Y cursor positions.
         targetpospertrial (ntrials, 2): X and Y pos of target.

    Returns:
        (ntime, ntrials): Array including instantaneous direction to the target from the cursor [rad]

    '''
    ntime = trial_aligned_pos.shape[0]
    ntrials = trial_aligned_pos.shape[1]
    inst_target_dir = np.zeros((ntime, ntrials))*np.nan

    for itrial in range(ntrials):
        cursor_location = trial_aligned_pos[:, itrial,:]
        target_location = targetpospertrial[itrial, :]
        rel_target_angle, _ = get_relative_point_location(cursor_location, target_location)
        inst_target_dir[:, itrial] = rel_target_angle

    return inst_target_dir

def mean_fr_inst_dir(data, trial_aligned_pos, targetpos, data_binwidth, ntarget_directions, data_samplerate, cursor_samplerate):
    '''
    This function takes trial aligned neural data, cursor position, and target position then calculates
    the mean firing rate per target location. Each target location is the instantaneous target direction from the 
    current cursor position, and therefore has multiple values during a single trial. The target locaitons are determined by calling aopy.postproc.get_inst_target_dir. 
    The target directions are assumed to be evenly spaced around the origin and the 0'th target starts directly horizontal from the origin.
    This function is specific to the 2D case at the moment with X coordinates being the horizontal dimension and Y coordinates being the vertical dimension. 

    Args:
        data (ntime, nunit, ntrial): Trial aligned data
        trial_aligned_pos (ntime, ntrials, 2): Position of the cursor in X and Y coordinates.trial_aligned_pos[:,:,0] corresponds to the X cursor positions and trial_aligned_pos[:,:,1] to the Y cursor positions.
        targetpos (ntrial, 2): Target position for each trial. First column is x target position, second column is y target position.
        data_binwidth (float): Bin size for neural data and cursor position. Can not be smaller than allowed by cursor position sampling rate.
        ntarget_directions (float): Number of directions to bin instantaneous direction into.
        data_samplerate (int): Sampling rate for data
        cursor_samplerate (int): Sampling rate for cursor position

    Returns:
        (nunit, ntarget_directions): Average firing rate per unit per direction bin. [spikes/s]
        
    '''
    # Check that binwidths are not lower than allowed by sampling rate
    max_cursor_binwidth = 1/cursor_samplerate #[s/sample]
    if data_binwidth < max_cursor_binwidth:
        data_binwidth = max_cursor_binwidth
    
    ndatatime, nunit, ntrial = data.shape
    nbins = math.ceil(ndatatime/(data_samplerate*data_binwidth)) # the number of bins

    # Bin neural data and cursor pos for each trial to make them the same samplingrate
    binned_data = np.zeros((nbins, nunit, ntrial))*np.nan
    for itrial in range(ntrial):
        binned_data[:,:,itrial] = precondition.bin_spikes(data[:,:,itrial], data_samplerate, data_binwidth)

    # (use precondition.bin_spikes to get average value in each bin)
    binned_cursorxpos = precondition.bin_spikes(trial_aligned_pos[:,:,0], cursor_samplerate, data_binwidth)/cursor_samplerate
    binned_cursorypos = precondition.bin_spikes(trial_aligned_pos[:,:,1], cursor_samplerate, data_binwidth)/cursor_samplerate
    binned_cursorpos = np.concatenate((np.expand_dims(binned_cursorxpos,axis=2),np.expand_dims(binned_cursorypos,axis=2)), axis=2)

    # Get instantaneous target location for each cursor pos (ntime, ntrial)
    inst_target_dir = get_inst_target_dir(binned_cursorpos, targetpos)

    # Match the instantaneous direction to the correct target direction bin and ensure the bin
    # range starts entered on target 0 (directly horizontal from the origin)
    targetloc_binwidth = (2*np.pi)/ntarget_directions # [rad] Angular bin size of each target direction
    target_binid = 1 + (inst_target_dir-(targetloc_binwidth/2))//targetloc_binwidth
    target_binid[target_binid==ntarget_directions] = 0 #combine first and last bins

    # Average data and place into correct points in array
    mean_dir_fr = np.zeros((nunit, ntarget_directions))
    for iunit in range(nunit):
        for idir in range(ntarget_directions):
            temp_data = binned_data[:,iunit,:]
            mean_dir_fr[iunit, idir] = np.nanmean(temp_data[target_binid==idir])

    return mean_dir_fr

def sample_events(events, times, samplerate):
    '''
    Converts a list of events and timestamps to a matrix of events where
    each column is a different event and each row is a sample in time.
    For example, if we have events 'reward' and 'penalty', and we want them
    as separate rasters::

        >>> events = ["reward", "reward", "penalty", "reward"]
        >>> times = [0.3, 0.5, 0.7, 1.0]
        >>> samplerate = 10
        >>> frame_events, event_names = sample_events(events, times, samplerate)
        >>> print(frame_events)
        [[False, False],
         [False, False],
         [False, False],
         [False, True ],
         [False, False],
         [False, True ],
         [False, False],
         [ True, False],
         [False, False],
         [False, False],
         [False, True ]]
        >>> print(event_names)
        ["penalty", "reward"]

    Args:
        events (list): list of event names or numbers
        times (list): list of timestamps for each event
        samplerate (float): rate at which you want to sample the events

    Returns:
        tuple: tuple containing:
            | **frame_events (nt, n_events):** logical index of 'events' at the given sampling rate
            | **event_names (n_events):** list of event column names (sorted alphabetically)

    '''
    n_samples = round(times[-1]*samplerate) + 1
    unique_events = np.unique(events)
    frame_events = np.zeros((n_samples, len(unique_events)), dtype='bool')
    for idx_event in range(len(events)):
        unique_idx = unique_events == events[idx_event]
        event_time = times[idx_event]
        event_frame = round(event_time * samplerate)
        frame_events[event_frame,unique_idx] = True
        
    return frame_events, unique_events

def get_calibrated_eye_data(eye_data, coefficients):
    """
    Apply the least square fitting coefficients to segments of eye data
    
    Args:
        eye_data (nt, nch): data to calibrate. Typically 4 channels (left eye x, left eye y, right eye x, right eye y)
        coefficients (nch, 2): coefficients to use for calibration for each channel of data
        
    returns:
        (nt, nch) ndarray: calibrated data
    """    
    #caliberated_eye_data_segments = np.empty((num_time_points, num_dims))
    return eye_data * coefficients[:,0] + coefficients[:,1]

def _correct_hand_traj(bmi3d_task_data):
    '''
    This function removes hand position data points when the cursor is simultaneously stationary in all directions.
    These hand position data points are artifacts. 
        
    Args:
        exp_data (dict): BMI3D task data
    
    Returns:
        hand_position (nt, 3): Corrected hand position
    '''

    hand_position = bmi3d_task_data['manual_input']

    # Set hand position to np.nan if the cursor position doesn't update. This indicates an optitrack error moved the hand outside the boundary.
    bad_pt_mask = np.zeros(bmi3d_task_data['cursor'].shape, dtype=bool) 
    bad_pt_mask[1:,0] = (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,0] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,1] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,2]
    bad_pt_mask[1:,1] = (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,0] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,1] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,2]
    bad_pt_mask[1:,2] = (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,0] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,1] & (np.diff(bmi3d_task_data['cursor'], axis=0)==0)[:,2]
    hand_position[bad_pt_mask] = np.nan

    return hand_position
